﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using DevLeap.Linq.LinqToEntities.Model.FromDatabase;
using EF4SampleDB = DevLeap.Linq.LinqToEntities.Model.EF4SampleDB;
using System.Data.Objects;

namespace DevLeap.Linq.LinqToEntities.Model
{
public class Program
{
        static void Main()
        {
            // ManageCustomersAndOrdersWithIndependentAssociation();
            // ManageCustomersAndOrdersWithForeignKeyAssociation();
            InheritanceDemo();
            InvokeStoredProcedure();
            UseCUDStoredProcedures();
        }

        // Listing 8-3
        static void ManageCustomersAndOrdersWithIndependentAssociation()
        {
            NorthwindEntities nw = new NorthwindEntities();

            // Browse for orders of a specific customer
            var query =
                from o in nw.Orders
                where o.Customer.CustomerID == "ALFKI"
                select new
                {
                    o.OrderID,
                    o.OrderDate,
                    o.Customer.CustomerID
                };

            foreach (var o in query)
            {
                Console.WriteLine(o);
            }

            // Add a new order to a specific customer
            Order newOrder = new Order { 
                OrderDate = DateTime.Now,
                Customer = nw.Customers.Single(c => c.CustomerID == "ANATR"),
                ShipCountry = "Italy",
            };

            nw.Orders.AddObject(newOrder);
            nw.SaveChanges();
        }

        // Listing 8-4
        static void ManageCustomersAndOrdersWithForeignKeyAssociation()
        {
            NorthwindEntities nw = new NorthwindEntities();

            // Browse for orders of a specific customer
            var query =
                from o in nw.Orders
                where o.CustomerID == "ALFKI"
                select new
                {
                    o.OrderID,
                    o.OrderDate,
                    o.CustomerID
                };

            foreach (var o in query)
            {
                Console.WriteLine(o);
            }

            // Add a new order to a specific customer
            Order newOrder = new Order
            {
                OrderDate = DateTime.Now,
                CustomerID = "ANATR",
                ShipCountry = "Italy",
            };

            nw.Orders.AddObject(newOrder);
            nw.SaveChanges();
        }

        // Listing 8-6
        static void InheritanceDemo()
        {
            EF4SampleDB.EF4SampleDBEntities ctx = new EF4SampleDB.EF4SampleDBEntities();

            // Browse all the contacts based on their type
            foreach (var i in ctx.Contacts)
            {
                if (i is EF4SampleDB.Supplier)
                {
                    EF4SampleDB.Supplier s = (EF4SampleDB.Supplier)i;
                    Console.WriteLine("Supplier {0} has ID {1}, and has a reference of {2}.",
                        s.FullName, s.ContactID, s.Reference);
                }
                else if (i is EF4SampleDB.Customer)
                {
                    EF4SampleDB.Customer c = (EF4SampleDB.Customer)i;
                    Console.WriteLine("Customer {0} has ID {1}, with an income of {2} last year",
                        c.FullName, c.ContactID, c.LastYearIncomes);
                }
            }

            // Filter Customer entities only
            var customers = from c in ctx.Contacts.OfType<EF4SampleDB.Customer>()
                            select c;

            foreach (var c in customers)
            {
                Console.WriteLine("Customer {0} has ID {1}, with an income of {2} last year",
                    c.FullName, c.ContactID, c.LastYearIncomes);
            }
        }

        // Listing 8-8
        static void InvokeStoredProcedure()
        {
            EF4SampleDB.EF4SampleDBEntities ctx = new EF4SampleDB.EF4SampleDBEntities();
            ObjectResult<EF4SampleDB.GetCustomersByAddress_Result> result = ctx.GetCustomersByAddress("%Firenze%");
            
            foreach (var c in result)
            {
                Console.WriteLine("Customer {0} has ID {1}, with an income of {2} last year",
                    c.FullName, c.ContactID, c.LastYearIncomes);
            }
        }

        // Listing 8-10
        static void UseCUDStoredProcedures()
        {
            EF4SampleDB.EF4SampleDBEntities ctx = new EF4SampleDB.EF4SampleDBEntities();

            EF4SampleDB.Customer customerToAdd = new EF4SampleDB.Customer();
            customerToAdd.FullName = "Frank White";
            customerToAdd.EMail = "frank@email.com";
            customerToAdd.Address = "USA - Redmond";
            customerToAdd.LastYearIncomes = 300;
            customerToAdd.ShippingAddress.ShippingAddress = "Microsoft Way, 1";
            customerToAdd.ShippingAddress.ShippingCity = "Redmond, WA";
            customerToAdd.ShippingAddress.ShippingPostalCode = "98052";
            customerToAdd.ShippingAddress.ShippingCountry = "US";

            ctx.AddToContacts(customerToAdd);

            // Here it is executed the statement
            // exec [dbo].[AddCustomer] @FullName=N'Frank White',@EMail=N'frank@email.com',
            // @Address=N'USA - Redmond',@LastYearIncomes=300.0000,@ShippingAddress=N'Microsoft Way, 1',
            // @ShippingCity=N'Redmond, WA',@ShippingPostalCode=N'98052',@ShippingCountry=N'US'
            ctx.SaveChanges();

            Int32 newId = customerToAdd.ContactID;

            EF4SampleDB.Customer customerToUpdate = 
                (from c in ctx.Contacts.OfType<EF4SampleDB.Customer>()
                where c.ContactID == newId
                select c).FirstOrDefault();

            customerToUpdate.EMail = "frank@email.net";

            // Here it is executed the statement
            // exec [dbo].[UpdateCustomer] @ContactID=9,@FullName=N'Frank White',
            // @EMail=N'frank@email.net',@Address=N'USA - Redmond',@LastYearIncomes=300.0000,
            // @ShippingAddress=N'Microsoft Way, 1',@ShippingCity=N'Redmond, WA',
            // @ShippingPostalCode=N'98052',@ShippingCountry=N'US'
            ctx.SaveChanges();

            ctx.Contacts.DeleteObject(customerToUpdate);

            // Here it is executed the statement
            // exec [dbo].[DeleteCustomer] @ContactID=9
            ctx.SaveChanges();
        }
    }
}
