﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Linq;
using System.Threading;
using System.Globalization;

namespace DevLeap.Linq.LinqToObjects.Operators
{
    #region Types

    public enum Countries
    {
        USA,
        Italy,
    }

    public class Customer
    {
        public string Name;
        public string City;
        public Countries Country;
        public Order[] Orders;

        public override string ToString()
        {
            return String.Format("Name: {0} - City: {1} - Country: {2}",
                this.Name, this.City, this.Country);
        }

        private static volatile Customer empty;
        private static Object emptySyncLock = new Object();

        public static Customer Empty
        {
            get
            {
                // Multithreaded singleton pattern
                if (empty == null)
                {
                    lock (emptySyncLock)
                    {
                        if (empty == null)
                        {
                            empty = new Customer();
                            empty.Name = String.Empty;
                            empty.Country = Countries.Italy;
                            empty.City = String.Empty;
                            empty.Orders = (new List<Order>(Enumerable.Empty<Order>())).ToArray();
                        }
                    }
                }
                return (empty);
            }
        }
    }

    public class Order
    {
        public int IdOrder;
        public int Quantity;
        public bool Shipped;
        public string Month;
        public int IdProduct;

        public override string ToString()
        {
            return String.Format("IdOrder: {0} - IdProduct: {1} - Quantity: {2} - Shipped: {3} - Month: {4}",
                this.IdOrder, this.IdProduct, this.Quantity, this.Shipped, this.Month);
        }

        //public override bool Equals(object obj)
        //{
        //    if (!(obj is Order))
        //        return false;
        //    else
        //    {
        //        Order o = (Order)obj;
        //        return (o.IdOrder == this.IdOrder && o.IdProduct == this.IdProduct && o.Month == this.Month && o.Quantity == this.Quantity && o.Shipped == this.Shipped);
        //    }
        //}

        //public override int GetHashCode()
        //{
        //    return String.Format("{0}|{1}|{2}|{3}|{4}", this.IdOrder, this.IdProduct, this.Month, this.Quantity, this.Shipped).GetHashCode();
        //}
    }

    public class Product
    {
        public int IdProduct;
        public decimal Price;

        public override string ToString()
        {
            return String.Format("IdProduct: {0} - Price: {1}", this.IdProduct, this.Price);
        }

        public override bool Equals(object obj)
        {
            if (!(obj is Product))
                return false;
            else
            {
                Product p = (Product)obj;
                return (p.IdProduct == this.IdProduct && p.Price == this.Price);
            }
        }

        public override int GetHashCode()
        {
            return String.Format("{0}|{1}", this.IdProduct, this.Price).GetHashCode();
        }
    }

    #endregion

    class Program
    {
        private static Customer[] customers;
        private static Product[] products;

        static void LoadData()
        {
            customers = new Customer[] {
                new Customer {Name = "Paolo", City = "Brescia", Country = Countries.Italy, Orders = new Order[] {
                    new Order {IdOrder = 1, Quantity = 3, IdProduct = 1, Shipped = false, Month = "January"},
                    new Order {IdOrder = 2, Quantity = 5, IdProduct = 2, Shipped = true, Month = "May"}}},
                new Customer {Name = "Marco", City = "Torino", Country = Countries.Italy, Orders =  new Order[] {
                    new Order {IdOrder = 3, Quantity = 10, IdProduct = 1, Shipped = false, Month = "July"},
                    new Order {IdOrder = 4, Quantity = 20, IdProduct = 3, Shipped = true, Month = "December"}}},
                new Customer {Name = "James", City = "Dallas", Country = Countries.USA, Orders =  new Order[] {
                    new Order {IdOrder = 5, Quantity = 20, IdProduct = 3, Shipped = true, Month = "December"}}},
                new Customer {Name = "Frank", City = "Seattle", Country = Countries.USA, Orders =  new Order[] {
                    new Order {IdOrder = 6, Quantity = 20, IdProduct = 5, Shipped = false, Month = "July"}}}};

            products = new Product[] {
                new Product {IdProduct = 1, Price = 10 },
                new Product {IdProduct = 2, Price = 20 },
                new Product {IdProduct = 3, Price = 30 },
                new Product {IdProduct = 4, Price = 40 },
                new Product {IdProduct = 5, Price = 50 },
                new Product {IdProduct = 6, Price = 60 }};
        }

        static Program()
        {
            LoadData();
        }

        static void Main(string[] args)
        {
            // restrictionOperator();
            // restrictionOperatorWithIndex();
            // inMemoryPagingSample();
            // projectionWithIndex();
            // singleSelectOverInnerDataSource();
            // selectManySample();
            // selectManySampleInQueryExpression();
            // selectManySampleWithCustomSelector();
            // selectManySampleWithCustomSelectorInQueryExpression();
            // descendingOrdering();
            // multipleOrdering();
            // defaultComparerForCustomersOrders();
            // customMonthComparerForCustomersOrders();
            // reverseOperator();
            // reverseOperatorMixedSyntax();
            // customersGroupedByCountry();
            // customersGroupedByCountryQueryExpression();
            // customersGroupedByCountryCustomElementSelector();
            // customersGroupedByCountryCustomResultSelector();
            // customersGroupedByCountryCustomElementAndResultSelector();
            // joinOfProductsWithOrders();
            // joinOfProductsWithOrdersQueryExpression();
            // groupJoinBetweenProductsAndOrders();
            // joinIntoClause();
            // joinIntoClauseNestedSelectMany();
            // distinctSample();
            // distinctSampleOverQueryExpression();
            // unionSample();
            // unionSampleWithDuplicateItemsSemanticallyEquivalent();
            // intersectAndExceptSample();
            // intersectSampleOverQueryExpression();
            // countSample();
            zipSample();
            // sumSampleOverCustomersOrders();
            // sumSampleOverCustomersOrdersComplexQueryExpression();
            // minOrdersQuantity();
            // minOverWrongTypes();
            // maxOverCustomTypesWithSelector();
            // averageSamples();
            // averageToExtractCustomersWithAverageOrdersAmount();
            // aggregateToExtractMostExpensiveOrderForEachCustomer();
            // aggregateToExtractOrdersWithTheirTotalOrderedAmount();
            // aggregateToExtractMostExpensiveOrdersAndMonthForEachCustomer();
            // rangeSampleToFilterOrders();
            // factorial(4);
            // repeatSample();
            // emptySample();
            // anySample();
            // allSample();
            // containsSample();
            // takeSample();
            // takeWhileSample();
            // firstSample();
            // firstOrDefaultSample();
            // singleSample();
            // elementAtAndElementAtOrDefaultSample();
            // defaultIfEmptySample();
            // concatSample();
            // customWhereExtensionMethodForCustomers();
            // asEnumerableSample();
            // toListSample();
            // anotherToListSample();
            // toDictionarySample();
            // toDictionarySampleExecution();
            // toLookupSample();
        }

        //
        // Listing 3-1
        //
        static void restrictionOperator()
        {
            var expr = from c in customers
                       where c.Country == Countries.Italy
                       select new { c.Name, c.City };

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-2
        //
        static void restrictionOperatorWithIndex()
        {
            var expr = customers
                .Where((c, index) => (c.Country == Countries.Italy && index >= 1))
                .Select(c => c.Name);

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-3
        //
        static void inMemoryPagingSample()
        {
            int pageSize = 5;
            int currentPage = 10;

            var expr = customers
                .Where((c, index) => ((index >= ((currentPage - 1) * pageSize)) && index < (currentPage * pageSize)))
                .Select(c => c.Name);

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-4
        //
        static void projectionWithIndex()
        {
            var expr =
                customers
                .Select((c, index) => new { index, c.Name, c.Country });

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-5
        //
        static void singleSelectOverInnerDataSource()
        {
            var orders = customers
                .Where(c => c.Country == Countries.Italy)
                .Select(c => c.Orders);

            foreach (var item in orders)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-6
        //
        static void selectManySample()
        {
            var orders = customers
                .Where(c => c.Country == Countries.Italy)
                .SelectMany(c => c.Orders);

            foreach (var item in orders)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-7
        //
        static void selectManySampleInQueryExpression()
        {
            var orders = from c in customers
                         where c.Country == Countries.Italy
                         from o in c.Orders
                         select o;

            foreach (var item in orders)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-8
        //
        static void selectManySampleWithCustomSelector()
        {
            var items = customers
                .Where(c => c.Country == Countries.Italy)
                .SelectMany(c => c.Orders, (c, o) => new { o.Quantity, o.IdProduct });

            foreach (var item in items)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-9
        //
        static void selectManySampleWithCustomSelectorInQueryExpression()
        {
            var items = from c in customers
                        where c.Country == Countries.Italy
                        from o in c.Orders
                        select new { o.Quantity, o.IdProduct };

            foreach (var item in items)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-10
        //
        static void descendingOrdering()
        {
            var expr = from c in customers
                       where c.Country == Countries.Italy
                       orderby c.Name descending
                       select new { c.Name, c.City };

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-11
        //
        static void multipleOrdering()
        {
            var expr = from c in customers
                       where c.Country == Countries.Italy
                       orderby c.Name descending, c.City
                       select new { c.Name, c.City };

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-12
        //
        static void defaultComparerForCustomersOrders()
        {
            var expr = from c in customers
                       from o in c.Orders
                       orderby o.Month
                       select o;

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-13
        //
        static void customMonthComparerForCustomersOrders()
        {
            var orders = customers
                .SelectMany(c => c.Orders)
                .OrderBy(o => o.Month, new MonthComparer());

            foreach (var item in orders)
            {
                Console.WriteLine(item);
            }
        }

        private class MonthComparer : IComparer<string>
        {
            public int Compare(string x, string y)
            {
                DateTime xDate = System.DateTime.ParseExact(x, "MMMM", new System.Globalization.CultureInfo("en-US"));
                DateTime yDate = System.DateTime.ParseExact(y, "MMMM", new System.Globalization.CultureInfo("en-US"));
                return (Comparer<DateTime>.Default.Compare(xDate, yDate));
            }
        }

        //
        // Listing 3-14
        //
        static void reverseOperator()
        {
            var expr = customers
                .Where(c => c.Country == Countries.Italy)
                .OrderByDescending(c => c.Name)
                .ThenBy(c => c.City)
                .Select(c => new { c.Name, c.City })
                .Reverse();

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-15
        //
        static void reverseOperatorMixedSyntax()
        {
            var expr = (from c in customers
                        where c.Country == Countries.Italy
                        orderby c.Name descending, c.City
                        select new { c.Name, c.City }).Reverse();

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-16
        //
        static void customersGroupedByCountry()
        {
            var expr = customers
                .GroupBy(c => c.Country);

            foreach (IGrouping<Countries, Customer> customerGroup in expr)
            {
                Console.WriteLine("Country: {0}", customerGroup.Key);
                foreach (var item in customerGroup)
                {
                    Console.WriteLine("\t{0}", item);
                }
            }
        }

        //
        // Listing 3-17
        //
        static void customersGroupedByCountryQueryExpression()
        {
            var expr = from c in customers
                       group c by c.Country;

            foreach (IGrouping<Countries, Customer> customerGroup in expr)
            {
                Console.WriteLine("Country: {0}", customerGroup.Key);
                foreach (var item in customerGroup)
                {
                    Console.WriteLine("\t{0}", item);
                }
            }
        }

        //
        // Listing 3-18
        //
        static void customersGroupedByCountryCustomElementSelector()
        {
            var expr = customers
                .GroupBy(c => c.Country, c => c.Name);

            foreach (IGrouping<Countries, string> customerGroup in expr)
            {
                Console.WriteLine("Country: {0}", customerGroup.Key);
                foreach (var item in customerGroup)
                {
                    Console.WriteLine("\t{0}", item);
                }
            }
        }

        //
        // Listing 3-19
        //
        static void customersGroupedByCountryCustomResultSelector()
        {
            var expr = customers
              .GroupBy(c => c.Country,
                (k, c) => new { Key = k, Count = c.Count() });

            foreach (var group in expr)
            {
                Console.WriteLine("Key: {0} - Count: {1}", group.Key, group.Count);
            }
        }

        //
        // Listing 3-20
        //
        static void customersGroupedByCountryCustomElementAndResultSelector()
        {
            var expr = customers
                .GroupBy(
                    c => c.Country, // keySelector
                    c => new { OrdersCount = c.Orders.Count() }, // elementSelector
                    (key, elements) => new
                    { // resultSelector
                        Key = key,
                        Count = elements.Count(),
                        OrdersCount = elements
                            .Sum(item => item.OrdersCount)
                    });

            foreach (var group in expr)
            {
                Console.WriteLine("Key: {0} - Count: {1} - Orders Count: {2}",
                    group.Key, group.Count, group.OrdersCount);
            }
        }

        //
        // Listing 3-21
        //
        static void joinOfProductsWithOrders()
        {
            var expr = customers
                .SelectMany(c => c.Orders)
                .Join(products, o => o.IdProduct, p => p.IdProduct, (o, p) => new { o.Month, o.Shipped, p.IdProduct, p.Price });

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-22
        //
        static void joinOfProductsWithOrdersQueryExpression()
        {
            var expr = from c in customers
                       from o in c.Orders
                       join p in products
                            on o.IdProduct equals p.IdProduct
                       select new { o.Month, o.Shipped, p.IdProduct, p.Price };

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-24
        //
        static void groupJoinBetweenProductsAndOrders()
        {
            var expr = products
                .GroupJoin(
                    customers.SelectMany(c => c.Orders),
                    p => p.IdProduct,
                    o => o.IdProduct,
                    (p, orders) => new { p.IdProduct, Orders = orders });

            foreach (var item in expr)
            {
                Console.WriteLine("Product: {0}", item.IdProduct);
                foreach (var order in item.Orders)
                {
                    Console.WriteLine("\t{0}", order);
                }
            }
        }

        //
        // Listing 3-25
        //
        static void joinIntoClause()
        {
            var customersOrders =
                from c in customers
                from o in c.Orders
                select o;

            var expr =
                from p in products
                join o in customersOrders
                    on p.IdProduct equals o.IdProduct
                    into orders
                select new { p.IdProduct, Orders = orders };

            foreach (var item in expr)
            {
                Console.WriteLine("Product: {0}", item.IdProduct);
                foreach (var order in item.Orders)
                {
                    Console.WriteLine("\t{0}", order);
                }
            }
        }

        //
        // Listing 3-26
        //
        static void joinIntoClauseNestedSelectMany()
        {
            var expr = from p in products
                       join o in
                           (
                            from c in customers
                            from o in c.Orders
                            select o
                            ) on p.IdProduct equals o.IdProduct
                            into orders
                       select new { p.IdProduct, Orders = orders };

            foreach (var item in expr)
            {
                Console.WriteLine("Product: {0}", item.IdProduct);
                foreach (var order in item.Orders)
                {
                    Console.WriteLine("\t{0}", order);
                }
            }
        }

        //
        // Listing 3-27
        //
        static void distinctSample()
        {
            var expr = customers
                .SelectMany(c => c.Orders)
                .Join(products,
                        o => o.IdProduct,
                        p => p.IdProduct,
                        (o, p) => p)
                .Distinct();

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-28
        //
        static void distinctSampleOverQueryExpression()
        {
            var expr = (from c in customers
                        from o in c.Orders
                        join p in products
                            on o.IdProduct equals p.IdProduct
                        select p).Distinct();

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-29
        //
        static void unionSample()
        {
            Int32[] setOne = { 1, 5, 6, 9 };
            Int32[] setTwo = { 4, 5, 7, 11 };

            var union = setOne.Union(setTwo);
            foreach (var i in union)
            {
                Console.Write(i + ", ");
            }
        }

        //
        // Listing 3-30
        //
        static void unionSampleWithDuplicateItemsSemanticallyEquivalent()
        {
            // In order to test this sample code without value type semantic,
            // please comment GetHashCode and Equals of the Product type, 
            // on top of this code

            Console.WriteLine("Union");
            Product[] productSetOne = { 
                new Product {IdProduct = 46, Price = 1000 },
                new Product {IdProduct = 27, Price = 2000 },
                new Product {IdProduct = 14, Price = 500}};
            Product[] productSetTwo = { 
                new Product {IdProduct = 11, Price = 350 },
                new Product {IdProduct = 46, Price = 1000 }};

            var productsUnion = productSetOne.Union(productSetTwo);

            Console.WriteLine(productSetOne[0].GetHashCode());
            Console.WriteLine(productSetTwo[1].GetHashCode());

            Console.WriteLine(productSetOne[0].Equals(productSetTwo[1].GetHashCode()));

            foreach (var item in productsUnion)
            {
                Console.WriteLine(item);
            }
        }

        private struct ProductValueType
        {
            public int IdProduct;
            public decimal Price;

            public override string ToString()
            {
                return String.Format("IdProduct: {0} - Price: {1}", this.IdProduct, this.Price);
            }
        }

        //
        // Listing 3-31
        //
        static void intersectAndExceptSample()
        {
            // Use ProductValueType instead of Product
            // to test the value type behavior

            //ProductValueType[] productSetOne = { 
            //    new ProductValueType {IdProduct = 46, Price = 1000 },
            //    new ProductValueType {IdProduct = 27, Price = 2000 },
            //    new ProductValueType {IdProduct = 14, Price = 500}};
            //ProductValueType[] productSetTwo = { 
            //    new ProductValueType {IdProduct = 11, Price = 350 },
            //    new ProductValueType {IdProduct = 46, Price = 1000 }};

            Product[] productSetOne = { 
                new Product {IdProduct = 46, Price = 1000 },
                new Product {IdProduct = 27, Price = 2000 },
                new Product {IdProduct = 14, Price = 500}};
            Product[] productSetTwo = { 
                new Product {IdProduct = 11, Price = 350 },
                new Product {IdProduct = 46, Price = 1000 }};

            // Intersect
            Console.WriteLine("Intersect");
            var expr = productSetOne.Intersect(productSetTwo);

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }

            // Intersect
            Console.WriteLine("Except");
            expr = productSetOne.Except(productSetTwo);

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-32
        //
        static void intersectSampleOverQueryExpression()
        {
            var expr = (from c in customers
                        from o in c.Orders
                        join p in products on o.IdProduct equals p.IdProduct
                        where c.Country == Countries.Italy
                        select p)
                .Intersect(
                        from c in customers
                        from o in c.Orders
                        join p in products on o.IdProduct equals p.IdProduct
                        where c.Country == Countries.USA
                        select p);

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-33
        //
        static void zipSample()
        {
            Int32[] numbers = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 };
            DayOfWeek[] weekDays = { 
                DayOfWeek.Sunday, 
                DayOfWeek.Monday, 
                DayOfWeek.Tuesday, 
                DayOfWeek.Wednesday, 
                DayOfWeek.Thursday, 
                DayOfWeek.Friday, 
                DayOfWeek.Saturday};

            var weekDaysNumbers = numbers.Zip(weekDays, 
                (first, second) => first + " - " + second);

            foreach (var item in weekDaysNumbers)
                Console.WriteLine(item);
        }

        //
        // Listing 3-34
        //
        static void countSample()
        {
            var expr = from c in customers
                       select new { c.Name, c.Country, OrdersCount = c.Orders.Count() };

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-35
        //
        static void sumSampleOverCustomersOrders()
        {
            var customersOrders = from c in customers
                                  from o in c.Orders
                                  join p in products
                                      on o.IdProduct equals p.IdProduct
                                  select new { c.Name, OrderAmount = o.Quantity * p.Price };

            Console.WriteLine("Single orders");
            foreach (var item in customersOrders)
            {
                Console.WriteLine(item);
            }

            var expr = from c in customers
                       join o in customersOrders
                            on c.Name equals o.Name
                            into customersWithOrders
                       select new { c.Name, TotalAmount = customersWithOrders.Sum(o => o.OrderAmount) };

            Console.WriteLine("Aggregated orders");
            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-36
        //
        static void sumSampleOverCustomersOrdersComplexQueryExpression()
        {
            var expr = from c in customers
                       join o in
                           (from c in customers
                            from o in c.Orders
                            join p in products
                                on o.IdProduct equals p.IdProduct
                            select new { c.Name, OrderAmount = o.Quantity * p.Price })
                            on c.Name equals o.Name
                            into customersWithOrders
                       select new { c.Name, TotalAmount = customersWithOrders.Sum(o => o.OrderAmount) };

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-37
        //
        static void minOrdersQuantity()
        {
            var expr =
                (from c in customers
                 from o in c.Orders
                 select o.Quantity).Min();

            Console.WriteLine(expr);
        }

        //
        // Listing 3-38
        //
        static void minOverWrongTypes()
        {
            try
            {
                var expr =
                    (from c in customers
                     from o in c.Orders
                     select new { o.IdProduct, o.Quantity }).Min();

                Console.WriteLine(expr);
            }
            catch (ArgumentException ex)
            {
                Console.WriteLine(ex);
            }
        }

        //
        // Listing 3-39
        //
        static void maxOverCustomTypesWithSelector()
        {
            var expr = (from c in customers
                        from o in c.Orders
                        select new { o.IdProduct, o.Quantity }).Max(o => o.Quantity);

            Console.WriteLine(expr);
        }

        //
        // Listing 3-40
        //
        static void averageSamples()
        {
            var expr = (from p in products
                        select p.Price).Average();

            Console.WriteLine(expr);

            expr = (from p in products
                    select new { p.IdProduct, p.Price }).Average(p => p.Price);

            Console.WriteLine(expr);
        }

        //
        // Listing 3-41
        //
        static void averageToExtractCustomersWithAverageOrdersAmount()
        {
            var expr = from c in customers
                       join o in
                           (from c in customers
                            from o in c.Orders
                            join p in products
                                on o.IdProduct equals p.IdProduct
                            select new { c.Name, OrderAmount = o.Quantity * p.Price })
                            on c.Name equals o.Name
                            into customersWithOrders
                       select new { c.Name, AverageAmount = customersWithOrders.Average(o => o.OrderAmount) };

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-42
        //
        static void aggregateToExtractMostExpensiveOrderForEachCustomer()
        {
            var expr = from c in customers
                       join o in
                           (from c in customers
                            from o in c.Orders
                            join p in products on o.IdProduct equals p.IdProduct
                            select new { c.Name, o.IdProduct, OrderAmount = o.Quantity * p.Price })
                            on c.Name equals o.Name
                            into orders
                       select new
                       {
                           c.Name,
                           MaxOrderAmount =
                               orders
                               .Aggregate((t, s) => t.OrderAmount > s.OrderAmount ?
                                                    t : s)
                               .OrderAmount
                       };

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-43
        //
        static void aggregateToExtractOrdersWithTheirTotalOrderedAmount()
        {
            var expr = from p in products
                       join o in
                           (from c in customers
                            from o in c.Orders
                            join p in products on o.IdProduct equals p.IdProduct
                            select new { p.IdProduct, OrderAmount = o.Quantity * p.Price })
                       on p.IdProduct equals o.IdProduct into orders
                       select new
                       {
                           p.IdProduct,
                           TotalOrderedAmount =
                            orders
                            .Aggregate(0m, (a, o) => a += o.OrderAmount)
                       };

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-44
        //
        static void aggregateToExtractMostExpensiveOrdersAndMonthForEachCustomer()
        {
            var expr = from c in customers
                       join o in
                           (from c in customers
                            from o in c.Orders
                            join p in products on o.IdProduct equals p.IdProduct
                            select new { c.Name, o.IdProduct, o.Month, OrderAmount = o.Quantity * p.Price })
                       on c.Name equals o.Name into orders
                       select new
                       {
                           c.Name,
                           MaxOrder = orders
                                        .Aggregate(new { Amount = 0m, Month = String.Empty },
                                                    (t, s) => t.Amount > s.OrderAmount
                                                    ? t
                                                    : new { Amount = s.OrderAmount, Month = s.Month })
                       };

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-48
        //
        static void rangeSampleToFilterOrders()
        {
            var expr = Enumerable.Range(1, 6)
                .SelectMany(x => (
                    from o in
                        (
                            from c in customers
                            from o in c.Orders
                            select o)
                    where o.Month == new CultureInfo("en-US").DateTimeFormat.GetMonthName(x)
                    select new { o.Month, o.IdProduct }));

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-49
        //
        static void factorial(int number)
        {
            var factorial = Enumerable.Range(0, number + 1).Aggregate(0, (s, t) => t == 0 ? 1 : s *= t);

            Console.WriteLine(factorial);
        }

        //
        // Listing 3-50
        //
        static void repeatSample()
        {
            var expr = Enumerable.Repeat((from c in customers select c.Name), 2).SelectMany(x => x);

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-51
        //
        static void emptySample()
        {
            IEnumerable<Customer> customers = Enumerable.Empty<Customer>();

            foreach (var item in customers)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-52
        //
        static void anySample()
        {
            bool result =
                (from c in customers
                 from o in c.Orders
                 select o)
                .Any(o => o.IdProduct == 1);

            Console.WriteLine(result);

            result = Enumerable.Empty<Order>().Any();

            Console.WriteLine(result);
        }

        //
        // Listing 3-53
        //
        static void allSample()
        {
            bool result =
                (from c in customers
                 from o in c.Orders
                 select o)
                .All(o => o.Quantity > 0);

            Console.WriteLine(result);

            result = Enumerable.Empty<Order>().All(o => o.Quantity > 0);

            Console.WriteLine(result);
        }

        //
        // Listing 3-54
        //
        static void containsSample()
        {
            var orderOfProductOne = new Order { IdOrder = 1, Quantity = 3, IdProduct = 1, Shipped = false, Month = "January" };
            bool result = customers[0].Orders.Contains(orderOfProductOne);

            Console.WriteLine(result);
        }

        //
        // Listing 3-55
        //
        static void takeSample()
        {
            var topTwoCustomers = (from c in customers
                                   join o in
                                       (from c in customers
                                        from o in c.Orders
                                        join p in products
                                            on o.IdProduct equals p.IdProduct
                                        select new { c.Name, OrderAmount = o.Quantity * p.Price })
                                       on c.Name equals o.Name
                                       into customersWithOrders
                                   let TotalAmount = customersWithOrders.Sum(o => o.OrderAmount)
                                   orderby TotalAmount descending
                                   select new { c.Name, TotalAmount }).Take(2);

            foreach (var item in topTwoCustomers)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-56
        //
        static void takeWhileSample()
        {
            // globalAmount is the total amount for all the orders
            var globalAmount = (from c in customers
                                from o in c.Orders
                                join p in products on o.IdProduct equals p.IdProduct
                                select new { OrderAmount = o.Quantity * p.Price }).Sum(o => o.OrderAmount);

            var limitAmount = globalAmount * 0.8m;
            var aggregated = 0m;
            var topCustomers =
                (from c in customers
                 join o in
                     (
                          from c in customers
                          from o in c.Orders
                          join p in products
                                 on o.IdProduct equals p.IdProduct
                          select new { c.Name, OrderAmount = o.Quantity * p.Price }
                          ) on c.Name equals o.Name
                         into customersWithOrders
                 let TotalAmount = customersWithOrders.Sum(o => o.OrderAmount)
                 orderby TotalAmount descending
                 select new { c.Name, TotalAmount }
                )
                .TakeWhile(X =>
                {
                    bool result = aggregated < limitAmount;
                    aggregated += X.TotalAmount;
                    return result;
                });

            foreach (var item in topCustomers)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-57
        //
        static void firstSample()
        {
            var item = customers.First(c => c.Country == Countries.USA);

            Console.WriteLine(item);
        }

        //
        // Listing 3-58
        //
        static void firstOrDefaultSample()
        {
            var item = customers.FirstOrDefault(c => c.City == "Las Vegas");
            Console.WriteLine(item == null ? "null" : item.ToString()); // returns null

            IEnumerable<Customer> emptyCustomers = Enumerable.Empty<Customer>();
            item = emptyCustomers.FirstOrDefault(c => c.City == "Las Vegas");
            Console.WriteLine(item == null ? "null" : item.ToString()); // returns null
        }

        //
        // Listing 3-59
        //
        static void singleSample()
        {
            var item = products.Single(p => p.IdProduct == 1);
            Console.WriteLine(item == null ? "null" : item.ToString()); // returns Product 1

            try
            {
                item = products.Single();
                Console.WriteLine(item == null ? "null" : item.ToString()); // InvalidOperationException
            }
            catch (InvalidOperationException ex)
            {
                Console.WriteLine(ex);
            }

            try
            {
                IEnumerable<Product> emptyProducts = Enumerable.Empty<Product>();
                item = emptyProducts.Single(p => p.IdProduct == 1);
                Console.WriteLine(item == null ? "null" : item.ToString()); // InvalidOperationException
            }
            catch (InvalidOperationException ex)
            {
                Console.WriteLine(ex);
            }
        }


        //
        // Listing 3-60
        //
        static void elementAtAndElementAtOrDefaultSample()
        {
            var item = products.ElementAt(2);
            Console.WriteLine(item == null ? "null" : item.ToString()); // returns Product 2

            item = Enumerable.Empty<Product>().ElementAtOrDefault(6);
            Console.WriteLine(item == null ? "null" : item.ToString()); // returns null

            item = products.ElementAtOrDefault(6);
            Console.WriteLine(item == null ? "null" : item.ToString()); // returns null
        }

        //
        // Listing 3-61
        //
        static void defaultIfEmptySample()
        {
            var expr = customers.DefaultIfEmpty();

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }

            IEnumerable<Customer> customersEmpty = Enumerable.Empty<Customer>().DefaultIfEmpty(Customer.Empty);

            foreach (var item in customersEmpty)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-62
        //
        static void concatSample()
        {
            var italianCustomers = from c in customers
                                   where c.Country == Countries.Italy
                                   select c;

            var americanCustomers = from c in customers
                                    where c.Country == Countries.USA
                                    select c;

            var expr = italianCustomers.Concat(americanCustomers);

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-64
        //
        static void customWhereExtensionMethodForCustomers()
        {
            Customers customersList = new Customers(customers);

            var expr = from c in customersList
                       where c.Country == Countries.Italy
                       select c;

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-65
        //
        static void asEnumerableSample()
        {
            Customers customersList = new Customers(customers);

            var expr = from c in customersList.AsEnumerable()
                       where c.City == "Brescia"
                       select c;

            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 3-66
        //
        static void toListSample()
        {
            List<Customer> customersList = new List<Customer>(customers);

            var expr = (from c in customersList
                        where c.Country == Countries.Italy
                        select c).ToList();

            Console.WriteLine("\nItems after query definition: {0}", expr.Count());
            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }

            customersList.Add(
                new Customer
                {
                    Name = "Luca",
                    City = "Firenze",
                    Country = Countries.Italy,
                    Orders = new Order[] {
                    new Order {Quantity = 3, IdProduct = 3, Shipped = false, Month = "March"}}
                });

            Console.WriteLine("\nItems after source QueryExpression modification: {0}", expr.Count());
            foreach (var item in expr)
            {
                Console.WriteLine(item);
            }

            Console.WriteLine("\nItems in customersList: {0}", customersList.Count);
        }

        //
        // Listing 3-67
        //
        static void anotherToListSample()
        {
            var productsQuery = (from p in products
                                 where p.Price >= 30
                                 select p).ToList();

            var ordersQueryWithProducts =
                from c in customers
                from o in c.Orders
                join p in productsQuery on o.IdProduct equals p.IdProduct
                select new { p.IdProduct, o.Quantity, p.Price, TotalAmount = o.Quantity * p.Price };

            foreach (var order in ordersQueryWithProducts)
            {
                Console.WriteLine(order);
            }
        }

        //
        // Listing 3-68
        //
        static void toDictionarySample()
        {
            var customersDictionary = customers
                .ToDictionary(c => c.Name, c => new { c.Name, c.City });

            foreach (var item in customersDictionary)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Unnumbered sample
        //
        static void toDictionarySampleExecution()
        {
            List<Customer> customersList = new List<Customer>(customers);

            var customersDictionary = customersList
                .ToDictionary(c => c.Name);

            Console.WriteLine("CustomersDictionary before source list modification");
            foreach (var item in customersDictionary)
            {
                Console.WriteLine("Key: {0} - Name: {1} - HashCode: {2}", item.Key, item.Value.Name, item.Value.GetHashCode());
            }

            customersList.Add(
                new Customer
                {
                    Name = "Luca",
                    City = "Firenze",
                    Country = Countries.Italy,
                    Orders = new Order[] {
                    new Order {Quantity = 3, IdProduct = 3, Shipped = false, Month = "March"}}
                });

            customersList[1].Name = "Changed";

            Console.WriteLine("CustomersDictionary after source list modification");
            foreach (var item in customersDictionary)
            {
                Console.WriteLine("Key: {0} - Name: {1} - HashCode: {2}", item.Key, item.Value.Name, item.Value.GetHashCode());
            }

            Console.WriteLine("Source customersList");
            foreach (var item in customersList)
            {
                Console.WriteLine("Name: {0} - HashCode: {1}", item.Name, item.GetHashCode());
            }

            var customersQuery =
                from c in customersList
                select c;

            Console.WriteLine("Query Expression over source customersList");
            foreach (var item in customersQuery)
            {
                Console.WriteLine("Name: {0} - HashCode: {1}", item.Name, item.GetHashCode());
            }
        }

        //
        // Listing 3-69
        //
        static void toLookupSample()
        {
            var ordersByProduct = (from c in customers
                                   from o in c.Orders
                                   select o).ToLookup(o => o.IdProduct);

            Console.WriteLine("\n\nNumber of orders for Product 1: {0}\n", ordersByProduct[1].Count());

            foreach (var product in ordersByProduct)
            {
                Console.WriteLine("Product: {0}", product.Key);
                foreach (var order in product)
                {
                    Console.WriteLine("  {0}", order);
                }
            }
        }
    }

    //
    // Listing 3-63
    //
    public sealed class Customers : List<Customer>
    {
        public Customers(IEnumerable<Customer> items)
            : base(items)
        {
        }
    }

    public static class CustomersEnumerable
    {
        public static IEnumerable<Customer> Where(this Customers source, Func<Customer, bool> predicate)
        {
            CustomerWhereIterator result = new CustomerWhereIterator(source, predicate);
            return result;
        }

        private class CustomerWhereIterator : IEnumerable<Customer>, IEnumerable
        {
            private Func<Customer, bool> _predicate;
            private Customers _customers;

            public CustomerWhereIterator(Customers source, Func<Customer, bool> predicate)
            {
                this._customers = source;
                this._predicate = predicate;
            }

            public IEnumerator<Customer> GetEnumerator()
            {
                foreach (var item in this._customers)
                {
                    Console.WriteLine("Custom Where implementation called");
                    if (this._predicate(item))
                    {
                        Console.WriteLine("Item matched predicate!");
                        yield return item;
                    }
                }
            }

            IEnumerator IEnumerable.GetEnumerator()
            {
                return this.GetEnumerator();
            }
        }
    }
}
