﻿Imports System.Runtime.CompilerServices

Module Program

    Private customers As Customer()
    Private products As Product()

    Sub Main()

        LoadData()
        ' joinOfProductsWithOrdersQueryExpression()
        ' averagePriceOfProductsOrderedByCustomers()
        ' averagePriceOfProductsOrderedByCustomersQueryExpression()
        useStandardDeviation()
        Console.ReadLine()

    End Sub

    Sub LoadData()

        customers = New Customer() { _
            New Customer With {.Name = "Paolo", .City = "Brescia", .Country = Countries.Italy, .Orders = New Order() { _
                New Order With {.IdOrder = 1, .Quantity = 3, .IdProduct = 1, .Shipped = False, .Month = "January"}, _
                New Order With {.IdOrder = 2, .Quantity = 5, .IdProduct = 2, .Shipped = True, .Month = "May"}}}, _
            New Customer With {.Name = "Marco", .City = "Torino", .Country = Countries.Italy, .Orders = New Order() { _
                New Order With {.IdOrder = 3, .Quantity = 10, .IdProduct = 1, .Shipped = False, .Month = "July"}, _
                New Order With {.IdOrder = 4, .Quantity = 20, .IdProduct = 3, .Shipped = True, .Month = "December"}}}, _
            New Customer With {.Name = "James", .City = "Dallas", .Country = Countries.USA, .Orders = New Order() { _
                New Order With {.IdOrder = 5, .Quantity = 20, .IdProduct = 3, .Shipped = True, .Month = "December"}}}, _
            New Customer With {.Name = "Frank", .City = "Seattle", .Country = Countries.USA, .Orders = New Order() { _
                New Order With {.IdOrder = 6, .Quantity = 20, .IdProduct = 5, .Shipped = False, .Month = "July"}}}}

        products = New Product() { _
            New Product With {.IdProduct = 1, .Price = 10}, _
            New Product With {.IdProduct = 2, .Price = 20}, _
            New Product With {.IdProduct = 3, .Price = 30}, _
            New Product With {.IdProduct = 4, .Price = 40}, _
            New Product With {.IdProduct = 5, .Price = 50}, _
            New Product With {.IdProduct = 6, .Price = 60}}

    End Sub

    '
    ' Listing 3-23
    '
    Sub joinOfProductsWithOrdersQueryExpression()

        Dim expr = _
            From c In customers _
            From o In c.Orders _
            Join p In products _
                On o.IdProduct Equals p.IdProduct _
            Select o.Month, o.Shipped, p.IdProduct, p.Price

        For Each item In expr
            Console.WriteLine(item)
        Next

    End Sub

    '
    ' Listing 3-45
    '
    Sub averagePriceOfProductsOrderedByCustomers()

        Dim productsOrdered = _
            From c In customers _
            From o In c.Orders _
                Join p In products _
                On o.IdProduct Equals p.IdProduct _
            Select p

        Dim expr = Aggregate p In productsOrdered _
            Into Average(p.Price)

        Console.WriteLine(expr)

    End Sub

    '
    ' Listing 3-46
    '
    Sub averagePriceOfProductsOrderedByCustomersQueryExpression()

        Dim expr = Aggregate p In ( _
            From c In customers _
            From o In c.Orders _
                Join p In products _
                On o.IdProduct Equals p.IdProduct _
            Select p) _
            Into Average(p.Price)

        Console.WriteLine(expr)

    End Sub

    <Extension()> _
    Function StandardDeviation( _
        ByVal source As IEnumerable(Of Double)) As Double

        If source Is Nothing Then
            Throw New ArgumentNullException("source")
        End If

        If source.Count = 0 Then
            Throw New InvalidOperationException("Cannot compute Standard Deviation for an empty set.")
        End If

        Dim avg = Aggregate v In source Into Average(v)
        Dim accumulator As Double = 0

        For Each x In source
            accumulator += (x - avg) ^ 2
        Next

        Return Math.Sqrt(accumulator / (source.Count))

    End Function

    <Extension()> _
    Function StandardDeviation(Of TSource)( _
            ByVal source As IEnumerable(Of TSource), _
            ByVal selector As Func(Of TSource, Double)) As Double

        Return (From element In source Select _
            selector(element)).StandardDeviation()

    End Function

    '
    ' Listing 3-47
    '
    Sub useStandardDeviation()

        Dim expr = Aggregate p In products _
           Into StandardDeviation(p.Price)

        Console.WriteLine(expr)

    End Sub

End Module

Public Enum Countries
    USA
    Italy
End Enum

Public Class Customer

    Public Name As String
    Public City As String
    Public Country As Countries
    Public Orders As Order()

    Public Overrides Function ToString() As String
        Return String.Format("Name: {0} - City: {1} - Country: {2}", _
            Me.Name, Me.City, Me.Country)
    End Function

End Class

Public Class Order

    Public IdOrder As Integer
    Public Quantity As Integer
    Public Shipped As Boolean
    Public Month As String
    Public IdProduct As Integer

    Public Overrides Function ToString() As String
        Return String.Format("IdOrder: {0} - IdProduct: {1} - Quantity: {2} - Shipped: {3} - Month: {4}", _
            Me.IdOrder, Me.IdProduct, Me.Quantity, Me.Shipped, Me.Month)
    End Function

End Class

Public Class Product

    Public IdProduct As Integer
    Public Price As Decimal

    Public Overrides Function ToString() As String
        Return String.Format("IdProduct: {0} - Price: {1}", _
            Me.IdProduct, Me.Price)
    End Function

End Class
