﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Data.Linq;

namespace DevLeap.Linq.LinqToSql {
    static class Helper {
        public static void DumpChanges(ChangeSet changeSet) {
            if (changeSet.Deletes.Count > 0) {
                Console.WriteLine("** DELETES **");
                foreach (var del in changeSet.Deletes) {
                    Console.WriteLine(Dump(del));
                }
            }
            if (changeSet.Updates.Count > 0) {
                Console.WriteLine("** UPDATES **");
                foreach (var upd in changeSet.Updates) {
                    Console.WriteLine(Dump(upd));
                }
            }
            if (changeSet.Inserts.Count > 0) {
                Console.WriteLine("** INSERTS **");
                foreach (var ins in changeSet.Inserts) {
                    Console.WriteLine(Dump(ins));
                }
            }
            Console.WriteLine(changeSet);
        }
        public static string Dump(this object data) {
            if (data is Customer) {
                Customer customer = (Customer) data;
                return String.Format(
                    "CustomerID={0}, CompanyName={1}", 
                    customer.CustomerID, customer.CompanyName);
            }
            else if (data is Order) {
                Order order = (Order) data;
                return String.Format(
                    "OrderID={0}, CustomerID={1}",
                    order.OrderID, order.CustomerID);
            }
            else if (data is Order_Detail) {
                Order_Detail detail = (Order_Detail) data;
                return String.Format(
                    "OrderID={0}, ProductID={1}, Quantity={2}",
                    detail.OrderID, detail.ProductID, detail.Quantity);
            }
            else if (data is Employee) {
                Employee employee = (Employee) data;
                return String.Format(
                    "EmployeeID={0}, FirstName={1}, LastName={2}, ReportsTo={3}",
                    employee.EmployeeID, employee.FirstName, employee.LastName, employee.ReportsTo);
            }
            else {
                throw new NotSupportedException(
                    String.Format(
                        "Dump is not supported on {0}",
                        data.GetType().FullName));
            }
        }
    }
}
