﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Xml.Linq;
using System.Linq;
using System.Xml;
using System.Xml.Schema;
using System.IO;
using System.Xml.Serialization;
using System.Xml.Xsl;
using System.Xml.XPath;
using System.Collections;
using System.Runtime.Serialization;

namespace DevLeap.Linq.LinqToXml.CSLinqToXml
{

    // NOTES: 
    // - The Listing 13-5 [An XML instance to search with LINQ to XML] is illustrative only
    // - For Listing 13-17 [Source XML with a list of customers] see the sourceCustomersForXslt.xml instance file
    // - For Listing 13-18 [Destination XML with a list of customers transformed] see the destinationCustomersFromXslt.xml instance file
    // - For Listing 13-19 [XSLT to transform XML from Listing 13-17 to Listing 13-18] see the customerFromSourceToDestination.xslt file

    // - The Listing 13-22 [A LINQ query over XML, based on an XSD typed approach] is illustrative only
    // - For Listing 13-23 [An XML Schema Definition for our sample list of customers] see the Customer.xsd file

    #region Support code

    public enum Countries
    {
        USA,
        Italy,
    }

    //
    // Listing 13-14
    //
    public class Customer
    {
        public string Name;
        public string City;
        public Countries Country;
        public Order[] Orders;

        public override string ToString()
        {
            return String.Format("Name: {0} - City: {1} - Country: {2}",
                this.Name, this.City, this.Country);
        }

        private static volatile Customer empty;
        private static Object emptySyncLock = new Object();

        public static Customer Empty
        {
            get
            {
                // Multithreaded singleton pattern
                if (empty == null)
                {
                    lock (emptySyncLock)
                    {
                        if (empty == null)
                        {
                            empty = new Customer();
                            empty.Name = String.Empty;
                            empty.Country = Countries.Italy;
                            empty.City = String.Empty;
                            empty.Orders = (new List<Order>(Enumerable.Empty<Order>())).ToArray();
                        }
                    }
                }
                return (empty);
            }
        }
    }

    public class Order
    {
        public int IdOrder;
        public int Quantity;
        public bool Shipped;
        public string Month;
        public int IdProduct;

        public override string ToString()
        {
            return String.Format("IdOrder: {0} - IdProduct: {1} - Quantity: {2} - Shipped: {3} - Month: {4}",
                this.IdOrder, this.IdProduct, this.Quantity, this.Shipped, this.Month);
        }

        public override bool Equals(object obj)
        {
            if (!(obj is Order))
                return false;
            else
            {
                Order o = (Order)obj;
                return (o.IdProduct == this.IdProduct && o.Month == this.Month && o.Quantity == this.Quantity && o.Shipped == this.Shipped);
            }
        }

        public override int GetHashCode()
        {
            return String.Format("{0}|{1}|{2}|{3}", this.IdProduct, this.Month, this.Quantity, this.Shipped).GetHashCode();
        }
    }

    public class Product
    {
        public int IdProduct;
        public decimal Price;

        public override string ToString()
        {
            return String.Format("{0} - {1}", this.IdProduct, this.Price);
        }
    }

    #endregion

    class Program
    {
        #region Support code

        private static Customer[] customers;
        private static Product[] products;

        static void LoadData()
        {
            customers = new Customer[] {
                new Customer {Name = "Paolo", City = "Brescia", Country = Countries.Italy, Orders = new Order[] {
                    new Order {IdOrder = 1, Quantity = 3, IdProduct = 1, Shipped = false, Month = "January"},
                    new Order {IdOrder = 2,Quantity = 5, IdProduct = 2, Shipped = true, Month = "May"}}},
                new Customer {Name = "Marco", City = "Torino", Country = Countries.Italy, Orders =  new Order[] {
                    new Order {IdOrder = 3,Quantity = 10, IdProduct = 1, Shipped = false, Month = "July"},
                    new Order {IdOrder = 4,Quantity = 20, IdProduct = 3, Shipped = true, Month = "December"}}},
                new Customer {Name = "James", City = "Dallas", Country = Countries.USA, Orders =  new Order[] {
                    new Order {IdOrder = 5,Quantity = 20, IdProduct = 3, Shipped = true, Month = "December"}}},
                new Customer {Name = "Frank", City = "Seattle", Country = Countries.USA, Orders =  new Order[] {
                    new Order {IdOrder = 6,Quantity = 20, IdProduct = 5, Shipped = false, Month = "July"}}}};

            products = new Product[] {
                new Product {IdProduct = 1, Price = 10 },
                new Product {IdProduct = 2, Price = 20 },
                new Product {IdProduct = 3, Price = 30 },
                new Product {IdProduct = 4, Price = 40 },
                new Product {IdProduct = 5, Price = 50 },
                new Product {IdProduct = 6, Price = 60 }};
        }

        static Program()
        {
            LoadData();
        }

        #endregion

        static void Main(string[] args)
        {
            // =====================================================
            // Chapter 12 Samples
            // =====================================================

            //sampleXmlDOMConstruction();
            //sampleXmlFunctionalConstruction();
            //xelementConstruction();
            //xmlelementConstruction();
            //directCastingXElement();
            //textEncodingXElement();
            //nodeNameEncodingXElement();
            //xattributeConstruction();
            //xnodeMethodsSample();
            //functionalConstructionWithLinqToObjects();
            //namespaceHandlingWithDOM();
            //namespaceHandlingWithLinqToXml();
            //multipleNamespacesHandlingWithLinqToXml();
            //namespaceWithPrefixWithLinqToXml();
            //namespaceQuerying();
            //defaultAndCustomPrefixNamespaces();
            //xstreamingElementSample();
            //chunkingXmlReaderWithXStreamingElement();
            //wrongSyntaxChunkingXmlReaderWithXStreamingElement();
            //annotationsSample();
            //eventsSample();
            //xelementReplaceWithMethod();
            //xelementAttributeChangeMethods();

            // =====================================================
            // Chapter 13 Samples
            // =====================================================

            //attributeExtensionMethods();
            //elementExtensionMethod();
            //useXElementMethodsWithNamespaces();
            //browseDescendantsResult();
            //descendantNodesSample();
            //nodesBeforeAfterSelf();
            //deferredQueryEvaluationOverXml();
            //queryXmlToBuildObjects();
            //mixedLinqToXmlAndLinqToObjects();
            //mixedLinqToXmlAndLinqToObjectsWithLet();
            //buildXmlCustomers();
            //createEntitiesFromXml();
            //createEntitiesFromXmlUsingChunkingReader();
            //functionalConstructionAsXsltTransformation();
            //validateXmlUsingLinqToXmlExtensions();
            //xmlValidationOnEvents();
            //xsltTransformationOverXNodes();
            //xpathEvaluateSample();
            //searchXmlViaXPathSelectElements();
            //xpathSelectElementsMethods();
            //secureChunkReaderSample();
            //xelementSerializationWithDataContract();
            //complexTypeWithXElementSerializationWithDataContract();

            Console.ReadLine();
        }

        //
        // Listing 12-1
        //
        static void sampleXmlDOMConstruction()
        {
            // Create the XmlDocument
            XmlDocument customerDocument = new XmlDocument();

            // Define processing instruction and document element (root element)
            customerDocument.AppendChild(customerDocument.CreateProcessingInstruction("xml", "version='1.0' encoding='UTF-16' standalone='yes'"));
            customerDocument.AppendChild(customerDocument.CreateElement("customer"));
            customerDocument.DocumentElement.SetAttribute("id", "C01");

            // Create and add "firstName" child element to the document element
            XmlElement firstNameElement = customerDocument.CreateElement("firstName");
            firstNameElement.InnerText = "Paolo";
            customerDocument.DocumentElement.AppendChild(firstNameElement);

            // Create and add "lastName" child element to the document element
            XmlElement lastNameElement = customerDocument.CreateElement("lastName");
            lastNameElement.InnerText = "Pialorsi";
            customerDocument.DocumentElement.AppendChild(lastNameElement);

            // Create "addresses" element
            XmlElement addressesElement = customerDocument.CreateElement("addresses");

            // Create and add "email address" child element to the "addresses" element
            XmlElement emailAddressElement = customerDocument.CreateElement("address");
            emailAddressElement.SetAttribute("type", "email");
            emailAddressElement.InnerText = "paolo@devleap.it";
            addressesElement.AppendChild(emailAddressElement);

            // Create and add "url address" child element to the "addresses" element
            XmlElement urlAddressElement = customerDocument.CreateElement("address");
            urlAddressElement.SetAttribute("type", "url");
            urlAddressElement.InnerText = "http://www.devleap.it/";
            addressesElement.AppendChild(urlAddressElement);

            // Create and add "home address" child element to the "addresses" element
            XmlElement homeAddressElement = customerDocument.CreateElement("address");
            homeAddressElement.SetAttribute("type", "home");
            homeAddressElement.InnerText = "Brescia - Italy";
            addressesElement.AppendChild(homeAddressElement);

            // Add "addresses" child element to the document element
            customerDocument.DocumentElement.AppendChild(addressesElement);

            Console.WriteLine(customerDocument.OuterXml);

            Console.WriteLine();
        }

        //
        // Listing 12-2
        //
        static void sampleXmlFunctionalConstruction()
        {
            XDocument customer =
              new XDocument(
                new XDeclaration("1.0", "UTF-16", "yes"),
                new XElement("customer",
                  new XAttribute("id", "C01"),
                  new XElement("firstName", "Paolo"),
                  new XElement("lastName", "Pialorsi"),
                  new XElement("addresses",
                    new XElement("address",
                      new XAttribute("type", "email"),
                      "paolo@devleap.it"),
                    new XElement("address",
                      new XAttribute("type", "url"),
                      "http://www.devleap.it/"),
                    new XElement("address",
                      new XAttribute("type", "home"),
                      "Brescia - Italy"))));

            customer.Save(Console.Out);

            Console.WriteLine();

            foreach (XElement a in customer.Descendants("addresses").Elements("address"))
            {
                Console.WriteLine(a);
            }
        }

        //
        // Listing 12-4
        //
        static void xelementConstruction()
        {
            XElement tag = new XElement("customer",
                new XElement("firstName", "Paolo"));

            Console.WriteLine(tag);
            Console.WriteLine();
        }

        //
        // Listing 12-5
        //
        static void xmlelementConstruction()
        {

            XmlDocument customerDocument = new XmlDocument();
            XmlElement customerElement = customerDocument.CreateElement("customer");
            XmlElement firstNameElement = customerDocument.CreateElement("firstName");
            firstNameElement.InnerText = "Paolo";
            customerElement.AppendChild(firstNameElement);
            customerDocument.AppendChild(customerElement);

            customerDocument.Save(Console.Out);
            Console.WriteLine();
        }

        //
        // Listing 12-7
        //
        static void directCastingXElement()
        {

            XElement order = new XElement("order",
                new XElement("quantity", 10),
                new XElement("price", 50),
                new XAttribute("idProdict", "P01"));

            Console.WriteLine(order);

            Decimal orderTotalAmount = (Decimal)order.Element("quantity") * (Decimal)order.Element("price");
            Console.WriteLine("Order total amount: {0}", orderTotalAmount);

            Console.ReadLine();
        }

        //
        // Listing 12-8
        //
        static void textEncodingXElement()
        {
            XElement notes = new XElement("notes", "Some special characters like & > < <div/> etc.");

            notes.Save(Console.Out);
            Console.WriteLine();
        }

        //
        // Listing 12-9
        //
        static void nodeNameEncodingXElement()
        {
            XElement s = new XElement(XmlConvert.EncodeNmToken("strange name!"),
                "My parent element has a strange name!");

            Console.WriteLine(s);
        }

        //
        // Listing 12-10
        //
        static void xattributeConstruction()
        {
            XElement customerTag = new XElement(
                "customer",
                new XAttribute("id", "C01"),
                "Paolo Pialorsi");

            Console.WriteLine(customerTag);
        }

        //
        // Listing 12-11
        //
        static void xnodeMethodsSample()
        {
            XElement customer = XElement.Load(@"..\..\customer.xml");
            XElement firstAddress = (customer.Descendants("addresses").Elements("address")).First();

            firstAddress.AddAfterSelf(
                new XElement("address",
                  new XAttribute("type", "IT-blog"),
                      "http://blogs.devleap.com/"),
                new XElement("address",
                  new XAttribute("type", "US-blog"),
                      "http://weblogs.asp.net/PaoloPia/"),
                      "franco", "nello");

            customer.Save(@"customer.xml");
            customer.Save(Console.Out);
        }

        //
        // Listing 12-12
        //
        static void functionalConstructionWithLinqToObjects()
        {
            XElement xmlCustomers = new XElement("customers",
                 from c in customers
                 where c.Country == Countries.Italy
                 select new XElement("customer",
                    new XAttribute("name", c.Name),
                    new XAttribute("city", c.City),
                    new XAttribute("country", c.Country)));

            xmlCustomers.Save(Console.Out);
        }

        //
        // Listing 12-14
        //
        static void namespaceHandlingWithDOM()
        {
            XmlDocument document = new XmlDocument();

            XmlElement customer = document.CreateElement("c", "customer", "http://schemas.devleap.com/Customer");
            document.AppendChild(customer);

            XmlElement firstName = document.CreateElement("c", "firstName", "http://schemas.devleap.com/Customer");
            firstName.InnerText = "Paolo Pialorsi";
            customer.AppendChild(firstName);

            document.Save(Console.Out);

            Console.WriteLine();
        }

        //
        // Listing 12-15
        //
        static void namespaceHandlingWithLinqToXml()
        {
            XNamespace ns = "http://schemas.devleap.com/Customer";
            XElement customer = new XElement(ns + "customer",
                new XAttribute("id", "C01"),
                new XElement(ns + "firstName", "Paolo"),
                new XElement(ns + "lastName", "Pialorsi"));

            customer.Save(Console.Out);
            Console.WriteLine();
        }

        //
        // Listing 12-17
        //
        static void multipleNamespacesHandlingWithLinqToXml()
        {
            XNamespace nsCustomer = "http://schemas.devleap.com/Customer";
            XNamespace nsAddress = "http://schemas.devleap.com/Address";

            XElement customer = new XElement(nsCustomer + "customer",
                new XAttribute("id", "C01"),
                new XElement(nsCustomer + "firstName", "Paolo"),
                new XElement(nsCustomer + "lastName", "Pialorsi"),
                new XElement(nsAddress + "addresses",
                    new XElement(nsAddress + "address",
                        new XAttribute("type", "email"),
                        "paolo@devleap.it"),
                    new XElement(nsAddress + "address",
                        new XAttribute("type", "home"),
                        "Brescia - Italy")));

            customer.Save(Console.Out);
            Console.WriteLine();

            Console.WriteLine(customer.Name);

        }

        //
        // Listing 12-18
        //
        static void namespaceWithPrefixWithLinqToXml()
        {
            XNamespace ns = "http://schemas.devleap.com/Customer";
            XElement customer = new XElement(ns + "customer",
                new XAttribute(XNamespace.Xmlns + "c", ns),
                new XAttribute("id", "C01"),
                new XElement(ns + "firstName", "Paolo"),
                new XElement(ns + "lastName", "Pialorsi"));

            customer.Save(Console.Out);
            Console.WriteLine();
        }

        //
        // Listing 12-19
        //
        static void namespaceQuerying()
        {
            XNamespace ns = "http://schemas.devleap.com/Customer";
            XElement customer = new XElement(ns + "customer",
                new XAttribute(XNamespace.Xmlns + "c", ns),
                new XAttribute("id", "C01"),
                new XElement(ns + "firstName", "Paolo"),
                new XElement(ns + "lastName", "Pialorsi"));

            customer.Save(Console.Out);

            Console.WriteLine();

            var namespaces =
                from n in customer.Attributes()
                where n.IsNamespaceDeclaration == true
                select n;

            foreach (var nsItem in namespaces)
            {
                Console.WriteLine("Value: {0}\nToString: {1}\nName: {2}",
                    nsItem.Value,
                    nsItem.ToString(),
                    nsItem.Name);
            }

            Console.WriteLine();
        }

        //
        // Listing 12-22
        //
        static void defaultAndCustomPrefixNamespaces()
        {
            XNamespace nsCustomer = "http://schemas.devleap.com/Customer";
            XNamespace nsAddress = "http://schemas.devleap.com/Address";
            XElement customer = new XElement(nsCustomer + "customer",
                new XAttribute("id", "C01"),
                new XElement(nsCustomer + "firstName", "Paolo"),
                new XElement(nsCustomer + "lastName", "Pialorsi"),
                new XElement(nsAddress + "address", "Brescia - Italy",
                    new XAttribute(XNamespace.Xmlns + "a", nsAddress)));

            customer.Save(Console.Out);
            Console.WriteLine();
        }

        //
        // Listing 12-23
        //
        static void xstreamingElementSample()
        {
            XStreamingElement xmlOrders =
                new XStreamingElement("orders",
                    from c in customers
                    from o in c.Orders
                    select new XStreamingElement("order",
                        new XStreamingElement("id", o.IdProduct),
                        new XStreamingElement("quantity", o.Quantity)));

            xmlOrders.Save(Console.Out);
        }

        //
        // Listing 12-24
        //
        static void chunkingXmlReaderWithXStreamingElement()
        {
            var customers = ChunkCustomers(@"..\..\CustomersWithOrders.xml");

            XStreamingElement xmlCustomers = new XStreamingElement("customers",
                customers);

            xmlCustomers.Save(@"..\..\CustomersWithOrdersOutput.xml");
        }

        //
        // Listing 12-25
        //
        static void wrongSyntaxChunkingXmlReaderWithXStreamingElement()
        {
            var customers = ChunkCustomers(@"..\..\CustomersWithOrders.xml");

            XStreamingElement xmlCustomers = new XStreamingElement("customers",
                    customers.ToList());

            xmlCustomers.Save(@"..\..\CustomersWithOrdersOutput.xml");
        }

        //
        // Listing 12-26
        //
        static void annotationsSample()
        {
            XElement customer = XElement.Load(@"..\..\customer.xml");

            CustomerAnnotation annotation = new CustomerAnnotation();
            annotation.Notes = "This is a good customer!";
            customer.AddAnnotation(annotation);

            customer.Save(Console.Out);

            annotation = customer.Annotation<CustomerAnnotation>();

            Console.WriteLine();
            Console.WriteLine(annotation.Notes);
        }

        /// <summary>
        /// One more complex example of Annotations, using derived annotation classes
        /// </summary>
        static void useAnnotationsOnXmlWithDerivedAnnotation()
        {
            XElement customer = XElement.Load(@"..\..\customer.xml");

            DerivedCustomerAnnotation annotation = new DerivedCustomerAnnotation();
            annotation.Notes = "This is a good customer!";
            annotation.IdRecord = 1;
            customer.AddAnnotation(annotation);

            customer.Save(Console.Out);

            annotation = (DerivedCustomerAnnotation)customer.Annotation<CustomerAnnotation>();

            Console.WriteLine();
            Console.WriteLine(annotation.Notes);
        }

        //
        // Listing 12-27
        //
        static void eventsSample()
        {
            XElement customer = XElement.Load(@"..\..\customer.xml");

            customer.Changing += (sender, e) =>
            {
                Console.WriteLine("=> Changing event raised");
                Console.WriteLine("\tChanging of type: {0}", e.ObjectChange);

                XObject x = sender as XObject;
                Console.WriteLine("\tChanging node of type: {0}", x.NodeType);

                XElement element = sender as XElement;
                if (element != null)
                {
                    Console.WriteLine("\tCurrent value: {0}", element.Value);
                }

                XAttribute attribute = sender as XAttribute;
                if (attribute != null)
                {
                    Console.WriteLine("\tCurrent value: {0}", attribute.Value);
                }
            };

            customer.Changed += (sender, e) =>
            {
                Console.WriteLine("=> Changed event raised");
                Console.WriteLine("\tChanged of type: {0}", e.ObjectChange);

                XObject x = sender as XObject;
                Console.WriteLine("\tChanged node of type: {0}", x.NodeType);

                XElement element = sender as XElement;
                if (element != null)
                {
                    Console.WriteLine("\tCurrent value: {0}", element.Value);
                }

                XAttribute attribute = sender as XAttribute;
                if (attribute != null)
                {
                    Console.WriteLine("\tCurrent value: {0}", attribute.Value);
                }
            };

            XElement emailAddress = customer.Descendants("address").First(a => a.Attribute("type").Value == "email");
            emailAddress.Value = "paolo@devleap.com";
            emailAddress.Attribute("type").Value = "externalEmail";
        }

        //
        // Listing 12-28
        //
        static void xelementReplaceWithMethod()
        {
            XElement customer = new XElement("customer",
                new XAttribute("id", "C01"),
                new XElement("firstName", "Paolo"),
                new XElement("lastName", "Pialorsi"));

            customer.Save(Console.Out);
            Console.WriteLine();

            customer.LastNode.ReplaceWith(
                new XElement("nickName", "PaoloPia"));

            customer.Save(Console.Out);
            Console.WriteLine();
        }

        //
        // Listing 12-29
        //
        static void xelementAttributeChangeMethods()
        {
            XElement customer = new XElement("customer",
                new XAttribute("id", "C01"));

            customer.Save(Console.Out);
            Console.WriteLine();

            customer.SetAttributeValue("id", "C02");

            customer.Save(Console.Out);
            Console.WriteLine();

            customer.ReplaceAttributes(new XAttribute("code", "CD01"));
            customer.SetAttributeValue("id", "C02");

            customer.SetElementValue("notes", "Notes about customer");
            customer.SetElementValue("notes", null);

            customer.Save(Console.Out);
            Console.WriteLine();
        }

        /// <summary>
        /// A preliminary example of events usage with LINQ to XML API
        /// </summary>
        static void useEvents()
        {
            XNamespace ns = "http://schemas.devleap.com/Customer";
            XElement customer = new XElement(ns + "customer",
                new XAttribute(XNamespace.Xmlns + "c", ns),
                new XAttribute("id", "C01"),
                new XElement(ns + "firstName", "Paolo"),
                new XElement(ns + "lastName", "Pialorsi"));

            customer.Changing += new EventHandler<XObjectChangeEventArgs>(customer_Changing);
            customer.Changed += new EventHandler<XObjectChangeEventArgs>(customer_Changed);

            customer.LastNode.ReplaceWith(
                new XElement(ns + "nickName", "PaoloPia"));

            customer.Save(Console.Out);
            Console.WriteLine();
        }

        static void customer_Changed(object sender, XObjectChangeEventArgs e)
        {
            Console.WriteLine(e.ObjectChange.ToString());
        }

        static void customer_Changing(object sender, XObjectChangeEventArgs e)
        {
            Console.WriteLine(e.ObjectChange.ToString());
        }

        //
        // Listing 13-1
        //
        static void attributeExtensionMethods()
        {
            XElement xmlCustomer = XElement.Load(@"..\..\customer.xml");

            Console.WriteLine(xmlCustomer);
            Console.WriteLine();

            Console.WriteLine("Attributes with name \"id\" count: {0}",
                xmlCustomer.Attributes().Where(a => a.Name == "id").Count());
            Console.WriteLine("\"id\" attribute value: {0}",
                xmlCustomer.Attribute("id").Value);

            Console.WriteLine();
        }


        //
        // Listing 13-2
        //
        static void elementExtensionMethod()
        {
            XElement xmlCustomers = new XElement("customers",
                 from c in customers
                 where c.Country == Countries.Italy
                 select new XElement("customer",
                    new XAttribute("name", c.Name),
                    new XAttribute("city", c.City),
                    new XAttribute("country", c.Country)));

            XElement element = xmlCustomers.Element("customer");
            Console.WriteLine(element);
            Console.WriteLine();

            // Listing 13-3 - Another sample LINQ to XML query based on the Elements extension method
            var elements = xmlCustomers.Elements();
            foreach (XElement e in elements)
            {
                Console.WriteLine(e);
            }
            Console.WriteLine();

            xmlCustomers.Save(Console.Out);
            Console.WriteLine();
        }

        //
        // Listing 13-4
        //
        static void useXElementMethodsWithNamespaces()
        {
            XNamespace ns = "http://schemas.devleap.com/Customers";
            XElement xmlCustomers = new XElement(ns + "customers",
                 from c in customers
                 where c.Country == Countries.Italy
                 select new XElement(ns + "customer",
                    new XAttribute("name", c.Name),
                    new XAttribute("city", c.City),
                    new XAttribute("country", c.Country)));

            XElement element = xmlCustomers.Element(ns + "customer");
            Console.WriteLine(element);
            Console.WriteLine();

            XElement firstElement = xmlCustomers.Elements().First();
            Console.WriteLine(firstElement);
            Console.WriteLine();

            xmlCustomers.Save(Console.Out);
            Console.WriteLine();
        }

        //
        // Listing 13-6
        //
        static void browseDescendantsResult()
        {
            XElement xmlCustomers = XElement.Load(@"..\..\Customers.xml");

            var descendantsOfXmlCustomers = xmlCustomers.Descendants();
            Console.WriteLine(descendantsOfXmlCustomers.Count());

            foreach (var descendant in descendantsOfXmlCustomers)
            {
                Console.WriteLine("=> Element: {0}\n{1}",
                    descendant.Name, descendant);
            }
        }

        //
        // Listing 13-7
        //
        static void descendantNodesSample()
        {
            XElement customer = XElement.Load(@"..\..\customer.xml");

            var descendantNodes = customer.DescendantNodes();
            Console.WriteLine("There are {0} descendant nodes",
                descendantNodes.Count());

            foreach (var descendant in descendantNodes)
            {
                Console.WriteLine("=> Node Type: {0}\n{1}",
                    descendant.NodeType, descendant);
            }
        }

        //
        // Listing 13-8
        //
        static void nodesBeforeAfterSelf()
        {
            XElement customer = XElement.Load(@"..\..\customer.xml");

            var firstAddress = customer.Element("addresses").FirstNode;
            var nodesAfterSelf = firstAddress.NodesAfterSelf();

            Console.WriteLine("Here is the first address:\n\t{0}",
                firstAddress);
            Console.WriteLine("There are {0} addresses after the first address",
                nodesAfterSelf.Count());
            foreach (var addressNode in nodesAfterSelf)
            {
                Console.WriteLine("\t{0}", addressNode);
            }

            Console.WriteLine();

            var lastAddress = customer.Element("addresses").LastNode;
            var nodesBeforeSelf = lastAddress.NodesBeforeSelf();

            Console.WriteLine("Here is the last address:\n\t{0}",
                lastAddress);
            Console.WriteLine("There are {0} addresses before the last address",
                nodesBeforeSelf.Count());
            foreach (var addressNode in nodesBeforeSelf)
            {
                Console.WriteLine("\t{0}", addressNode);
            }
        }

        //
        // Listing 13-09
        //
        static void deferredQueryEvaluationOverXml()
        {
            XElement xmlCustomers = new XElement("customers",
                 from c in customers
                 where c.Country == Countries.Italy
                 select new XElement("customer",
                    new XElement("name", c.Name),
                    new XElement("city", c.City),
                    new XElement("country", c.Country)));

            var cities = xmlCustomers.DescendantsAndSelf("city");
            var names = xmlCustomers.Elements("customer").Elements("name");

            Console.WriteLine("\nBefore XML source modification");
            foreach (var city in cities)
            {
                Console.WriteLine(city);
            }
            Console.WriteLine("---");
            foreach (var name in names)
            {
                Console.WriteLine(name);
            }

            //
            // Listing 13-10
            //
            xmlCustomers.Add(
                 from c in customers
                 where c.Country != Countries.Italy
                 select new XElement("customer",
                    new XElement("name", c.Name),
                    new XElement("city", c.City),
                    new XElement("country", c.Country)));

            Console.WriteLine("\nAfter XML source modification");
            foreach (var city in cities)
            {
                Console.WriteLine(city);
            }
            Console.WriteLine("---");
            foreach (var name in names)
            {
                Console.WriteLine(name);
            }
        }

        //
        // Listing 13-11
        //
        static void queryXmlToBuildObjects()
        {
            XElement xmlCustomers = new XElement("customers",
                 from c in customers
                 select new XElement("customer",
                    new XElement("name", c.Name),
                    new XAttribute("city", c.City),
                    new XAttribute("country", c.Country)));

            var customersFromXml =
                from c in xmlCustomers.Elements("customer")
                where (String)c.Attribute("country") == "Italy"
                orderby (String)c.Element("name")
                select new { Name = (String)c.Element("name"), City = (String)c.Attribute("city") };

            foreach (var customer in customersFromXml)
            {
                Console.WriteLine(customer);
            }
        }

        //
        // Listing 13-12
        //
        static void mixedLinqToXmlAndLinqToObjects()
        {
            XElement xmlCustomers = new XElement("customers",
                 from c in customers
                 select new XElement("customer",
                    new XAttribute("city", c.City),
                    new XElement("name", c.Name),
                    new XAttribute("country", c.Country)));

            var orders =
                from c in customers
                from o in c.Orders
                select new { c.Name, o.IdProduct, o.Quantity };

            var ordersWithCustomersFromXml =
                from c in xmlCustomers.Elements("customer")
                join o in orders
                on (String)c.Element("name") equals o.Name
                orderby (String)c.Element("name")
                select new
                {
                    Name = (String)c.Element("name"),
                    City = (String)c.Attribute("city"),
                    IdProduct = o.IdProduct,
                    Quantity = o.Quantity
                };

            foreach (var order in ordersWithCustomersFromXml)
            {
                Console.WriteLine(order);
            }
        }

        //
        // Listing 13-13
        //
        static void mixedLinqToXmlAndLinqToObjectsWithLet()
        {
            XElement xmlCustomers = new XElement("customers",
                 from c in customers
                 select new XElement("customer",
                    new XAttribute("city", c.City),
                    new XElement("name", c.Name),
                    new XAttribute("country", c.Country)));

            var orders =
                from c in customers
                from o in c.Orders
                select new { c.Name, o.IdProduct, o.Quantity };

            var ordersWithCustomersFromXml =
                from c in xmlCustomers.Elements("customer")
                let xName = (String)c.Element("name")
                let xCity = (String)c.Attribute("city")
                join o in orders
                on xName equals o.Name
                orderby xName
                select new
                {
                    Name = xName,
                    City = xCity,
                    IdProduct = o.IdProduct,
                    Quantity = o.Quantity
                };

            foreach (var order in ordersWithCustomersFromXml)
            {
                Console.WriteLine(order);
            }
        }

        //
        // Support code
        //
        static void buildXmlCustomers()
        {
            XElement sourceXmlCustomers = new XElement("customers",
                 from c in customers
                 where c.Country == Countries.Italy
                 select new XElement("customer",
                    new XAttribute("name", c.Name),
                    new XAttribute("city", c.City),
                    new XAttribute("country", c.Country),
                    new XElement("orders",
                        from o in c.Orders
                        select new XElement("order",
                            new XAttribute("id", o.IdOrder),
                            new XAttribute("idProduct", o.IdProduct),
                            new XAttribute("quantity", o.Quantity),
                            new XAttribute("shipped", o.Shipped),
                            new XAttribute("month", o.Month)))));

            sourceXmlCustomers.Save(@"..\..\customersWithOrdersDataSource.xml");
        }

        //
        // Listing 13-15
        //
        static void createEntitiesFromXml()
        {
            XElement xmlCustomers = XElement.Load(@"..\..\customersWithOrdersDataSource.xml");

            var customersWithOrders =
                from c in xmlCustomers.Elements("customer")
                select new Customer
                {
                    Name = (String)c.Attribute("name"),
                    City = (String)c.Attribute("city"),
                    Country = (Countries)Enum.Parse(typeof(Countries), (String)c.Attribute("country"), true),
                    Orders = (
                        from o in c.Descendants("order")
                        select new Order
                        {
                            IdOrder = (Int32)o.Attribute("id"),
                            IdProduct = (Int32)o.Attribute("idProduct"),
                            Quantity = (Int32)o.Attribute("quantity"),
                            Month = (String)o.Attribute("month"),
                            Shipped = (Boolean)o.Attribute("shipped")
                        }
                    ).ToArray()
                };

            foreach (var c in customersWithOrders)
            {
                Console.WriteLine(c);
                foreach (var o in c.Orders)
                {
                    Console.WriteLine("  {0}", o);
                }
            }
        }

        //
        // Listing 13-16
        //
        static void createEntitiesFromXmlUsingChunkingReader()
        {
            var xmlCustomers = ChunkedDataSourceReader(@"..\..\customersWithOrdersDataSource.xml", "customer");

            Console.WriteLine("Variable xmlCustomers defined");

            var customersWithOrders =
                from c in xmlCustomers
                select new Customer
                {
                    Name = (String)c.Attribute("name"),
                    City = (String)c.Attribute("city"),
                    Country = (Countries)Enum.Parse(typeof(Countries), (String)c.Attribute("country"), true),
                    Orders = (
                        from o in c.Descendants("order")
                        select new Order
                        {
                            IdOrder = (Int32)o.Attribute("id"),
                            IdProduct = (Int32)o.Attribute("idProduct"),
                            Quantity = (Int32)o.Attribute("quantity"),
                            Month = (String)o.Attribute("month"),
                            Shipped = (Boolean)o.Attribute("shipped")
                        }
                    ).ToArray()
                };

            Console.WriteLine("LINQ query expression defined");

            foreach (var c in customersWithOrders)
            {
                Console.WriteLine(c);
                foreach (var o in c.Orders)
                {
                    Console.WriteLine("  {0}", o);
                }
            }
        }

        static IEnumerable<XElement> ChunkedDataSourceReader(String uri, String chunkElement)
        {
            XmlReaderSettings settings = new XmlReaderSettings();
            settings.XmlResolver = new XmlSecureResolver(new XmlUrlResolver(), "c:\\sources");
            settings.MaxCharactersInDocument = (1024 * 1024) * 4; // MAX 4MB
            XmlReader xr = XmlReader.Create(uri, settings);

            while (xr.Read())
            {
                if ((xr.NodeType == XmlNodeType.Element)
                    && (xr.Name == chunkElement))
                {
                    // For debugging and demo purposes only
                    Console.WriteLine("Reading a chunk element from data source");

                    yield return XElement.ReadFrom(xr) as XElement;
                }
            }
        }

        //
        // Listing 13-20
        //
        static void functionalConstructionAsXsltTransformation()
        {
            XElement sourceXmlCustomers = new XElement("customers",
                 from c in customers
                 where c.Country == Countries.Italy
                 select new XElement("customer",
                    new XAttribute("name", c.Name),
                    new XAttribute("city", c.City),
                    new XAttribute("country", c.Country)));

            XNamespace ns = "http://schemas.devleap.com/Customers";
            XElement destinationXmlCustomers =
                new XElement(ns + "customers",
                    new XAttribute(XNamespace.Xmlns + "c", ns),
                    from c in sourceXmlCustomers.Elements("customer")
                    where c.Attribute("country").Value == "Italy"
                    select new XElement(ns + "customer",
                           new XElement(ns + "name", c.Attribute("name")),
                           new XElement(ns + "city", c.Attribute("city"))));

            destinationXmlCustomers.Save(Console.Out);
            Console.WriteLine();
        }

        /// <summary>
        /// Extra sample
        /// </summary>
        static void useAttributeMethods()
        {
            XElement xmlCustomers = new XElement("customers",
                 from c in customers
                 where c.Country == Countries.Italy
                 select new XElement("customer",
                    new XElement("name", c.Name),
                    new XElement("city", c.City),
                    new XElement("country", c.Country)));

            xmlCustomers.Save(Console.Out);

            Console.WriteLine(xmlCustomers.Descendants().Count());

            foreach (XNode a in xmlCustomers.DescendantsAndSelf().InDocumentOrder())
            {
                Console.WriteLine(a);
            }


            foreach (XAttribute a in xmlCustomers.Element("customer").Attributes("name"))
            {
                Console.WriteLine(a);
            }

            foreach (XElement e in xmlCustomers.Descendants())
            {
                Console.WriteLine(e);
            }
        }

        //
        // Listing 13-24
        //
        static void validateXmlUsingLinqToXmlExtensions()
        {
            XNamespace ns = "http://schemas.devleap.com/Customer";
            XDocument xmlCustomers = new XDocument(
                new XElement(ns + "customers",
                    from c in customers
                    select new XElement(ns + "customer",
                        new XAttribute("city", c.City),
                        new XAttribute("name", c.Name),
                        new XAttribute("country", c.Country))));

            xmlCustomers.Save(Console.Out);
            Console.WriteLine();

            //
            // Listing 13-25
            //
            XmlSchemaSet schemas = new XmlSchemaSet();
            schemas.Add(XmlSchema.Read(new StreamReader(@"..\..\customer.xsd"), null));

            xmlCustomers.Validate(schemas, xmlCustomers_validation);
        }

        static void xmlCustomers_validation(Object source, ValidationEventArgs args)
        {
            // In case of validation messages
            Console.WriteLine(args.Message);
        }

        //
        // Listing 13-26
        //
        static void xmlValidationOnEvents()
        {
            XNamespace ns = "http://schemas.devleap.com/Customer";
            XDocument xmlCustomers = new XDocument(
                new XElement(ns + "customers",
                    from c in customers
                    select new XElement(ns + "customer",
                        new XAttribute("city", c.City),
                        new XAttribute("name", c.Name),
                        new XAttribute("country", c.Country))));

            xmlCustomers.Save(Console.Out);
            Console.WriteLine();

            XmlSchemaSet schemas = new XmlSchemaSet();
            schemas.Add(XmlSchema.Read(new StreamReader(@"..\..\customer.xsd"), null));

            xmlCustomers.Validate(schemas, xmlCustomers_validation);

            // Subscribe to Changed event on the root element
            xmlCustomers.Changed += (sender, e) =>
            {
                xmlCustomers.Validate(schemas, (source, args) =>
                    {
                        if (args.Exception != null)
                            throw new InvalidOperationException("Operation on XML inconsistent with associated schema.", args.Exception);
                    });
            };

            // Changes the city attribute from the first customer (operation allowed)
            xmlCustomers.Element(ns + "customers")
                .Elements(ns + "customer").First()
                .SetAttributeValue("city", "Rodengo Saiano");

            // Removes the city attribute from the first customer (operation not allowed)
            xmlCustomers.Element(ns + "customers")
                .Elements(ns + "customer").First()
                .SetAttributeValue("city", null);
        }

        //
        // Listing 13-27
        //
        static void xsltTransformationOverXNodes()
        {
            XElement sourceXmlCustomers = new XElement("customers",
                 from c in customers
                 where c.Country == Countries.Italy
                 select new XElement("customer",
                    new XAttribute("name", c.Name),
                    new XAttribute("city", c.City),
                    new XAttribute("country", c.Country)));

            XslCompiledTransform xslt = new XslCompiledTransform();
            xslt.Load(@"..\..\customerFromSourceToDestination.xslt");

            Console.WriteLine("Source XML");
            sourceXmlCustomers.Save(Console.Out);
            Console.WriteLine("\n\n");

            Console.WriteLine("Destination XML");
            xslt.Transform(sourceXmlCustomers.CreateNavigator(), null, Console.Out);
            Console.WriteLine();
        }

        //
        // Listing 13-28
        //
        static void xpathEvaluateSample()
        {
            XElement xmlCustomers = new XElement("customers",
                 from c in customers
                 select new XElement("customer",
                    new XAttribute("name", c.Name),
                    new XAttribute("city", c.City),
                    new XAttribute("country", c.Country)));

            var result = (IEnumerable<Object>)xmlCustomers.XPathEvaluate("/customer[@country = 'Italy']/@name");

            foreach (var item in result)
            {
                Console.WriteLine(item);
            }

            Console.WriteLine();
        }

        /// <summary>
        /// Internal demo method
        /// </summary>
        static void searchXmlViaXPathSelectElements()
        {
            XElement xmlCustomers = new XElement("customers",
                 from c in customers
                 select new XElement("customer",
                    new XAttribute("name", c.Name),
                    new XAttribute("city", c.City),
                    new XAttribute("country", c.Country)));

            var result = xmlCustomers.XPathSelectElements("/customer[@country = 'Italy']");

            foreach (var item in result)
            {
                Console.WriteLine(item);
            }

            Console.WriteLine();
        }

        //
        // Listing 13-29
        //
        static void xpathSelectElementsMethods()
        {
            List<Customer> listCustomers = new List<Customer>(customers);

            XElement xmlCustomers = new XElement("customers",
                 from c in listCustomers
                 select new XElement("customer",
                    new XElement("name", c.Name),
                    new XAttribute("city", c.City),
                    new XAttribute("country", c.Country)));

            var orders =
                from c in listCustomers
                from o in c.Orders
                select new { c.Name, o.IdProduct, o.Quantity };

            var ordersOfItalianCustomersFromXml =
                from c in xmlCustomers.XPathSelectElements("/customer[@country = 'Italy']")
                let xName = (String)c.Element("name")
                let xCity = (String)c.Attribute("city")
                join o in orders
                on xName equals o.Name
                orderby xName
                select new
                {
                    Name = xName,
                    City = xCity,
                    IdProduct = o.IdProduct,
                    Quantity = o.Quantity
                };

            Console.WriteLine("First list");

            foreach (var order in ordersOfItalianCustomersFromXml)
            {
                Console.WriteLine(order);
            }

            xmlCustomers.Add(
                new XElement("customer",
                    new XElement("name", "Roberto"),
                    new XAttribute("city", "Firenze"),
                    new XAttribute("country", "Italy")));

            listCustomers.Add(new Customer
            {
                Name = "Roberto",
                City = "Firenze",
                Country = Countries.Italy,
                Orders = new Order[] {
                    new Order {Quantity = 10, IdProduct = 1, Shipped = false, Month = "July"},
                    new Order {Quantity = 20, IdProduct = 2, Shipped = true, Month = "Decembre"}}
            });

            Console.WriteLine("Second list");

            foreach (var order in ordersOfItalianCustomersFromXml)
            {
                Console.WriteLine(order);
            }
        }

        //
        // Listing 13-30
        //
        static void secureChunkReaderSample()
        {
            var customers = ChunkCustomers(@"..\..\CustomersWithOrders.xml");

            foreach (var c in customers)
            {
                Console.WriteLine(c);
            }
        }

        static IEnumerable<XElement> ChunkCustomers(String uri)
        {
            XmlReaderSettings settings = new XmlReaderSettings();
            settings.XmlResolver = new XmlSecureResolver(new XmlUrlResolver(), "c:\\sources");
            settings.MaxCharactersInDocument = (1024 * 1024) * 4; // MAX 4MB
            XmlReader xr = XmlReader.Create(uri, settings);

            while (xr.Read())
            {
                if ((xr.NodeType == XmlNodeType.Element)
                    && (xr.Name == "customer"))
                {
                    yield return XElement.ReadFrom(xr) as XElement;
                }
            }
        }

        //
        // Listing 13-31
        //
        static void xelementSerializationWithDataContract()
        {
            List<Customer> listCustomers = new List<Customer>(customers);

            XElement xmlCustomers = new XElement("customers",
                 from c in listCustomers
                 select new XElement("customer",
                    new XElement("name", c.Name),
                    new XAttribute("city", c.City),
                    new XAttribute("country", c.Country)));

            MemoryStream mem = new MemoryStream();
            DataContractSerializer dc = new DataContractSerializer(typeof(XElement));
            dc.WriteObject(mem, xmlCustomers);

            mem.Position = 0;

            StreamReader sr = new StreamReader(mem);
            Console.WriteLine(sr.ReadToEnd());
            sr.Close();

            mem = new MemoryStream();
            XmlSerializer xs = new XmlSerializer(typeof(XElement));
            xs.Serialize(mem, xmlCustomers);

            mem.Position = 0;

            sr = new StreamReader(mem);
            Console.WriteLine(sr.ReadToEnd());
        }

        //
        // Listing 13-32
        //
        static void complexTypeWithXElementSerializationWithDataContract()
        {
            List<Customer> listCustomers = new List<Customer>(customers);

            var customerQuery =
                 (from c in listCustomers
                  select new CustomerWithXmlData
                  {
                      Name = c.Name,
                      XmlData = new XElement("orders",
                             from o in c.Orders
                             select new XElement("order",
                                 new XAttribute("idProduct", o.IdProduct),
                                 new XAttribute("quantity", o.Quantity)))
                  }).ToList();

            MemoryStream mem = new MemoryStream();
            DataContractSerializer dc = new DataContractSerializer(customerQuery.GetType());
            dc.WriteObject(mem, customerQuery);

            mem.Position = 0;

            StreamReader sr = new StreamReader(mem);
            Console.WriteLine(sr.ReadToEnd());
            sr.Close();
        }
    }

    #region Support code

    public class CustomerAnnotation
    {
        private String _notes;

        public String Notes
        {
            get { return _notes; }
            set { _notes = value; }
        }
    }

    public class DerivedCustomerAnnotation : CustomerAnnotation
    {
        private Int32 _idRecord;

        public Int32 IdRecord
        {
            get { return _idRecord; }
            set { _idRecord = value; }
        }
    }

    [DataContract(Namespace = "http://schemas.devleap.com/CustomerWithXmlData")]
    public class CustomerWithXmlData
    {
        [DataMember]
        public XElement XmlData { get; set; }

        [DataMember]
        public String Name { get; set; }
    }

    #endregion
}
