﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Threading;
using System.Threading.Tasks;
using System.Xml.Linq;

namespace PLINQ {
    class Program {
        static void Main(string[] args) {
            
            // First run (cold cache)
            //StandardLinq();
            //SamplePLinq();
            //SamplePLinq_NoLatency();
            //SamplePLinq_StopAndGo();
            //SamplePLinq_Inverted();
            //SamplePLinq_Ordered();
            //SamplePLinq_Unordered();
            //SamplePLinq_ExceptionHandling();
            //SamplePLinq_Cancellation();
            //SamplePLinq_ForceParallel();
            //SamplePLinq_Sequential();
            //SamplePLinq_RaceCondition();

            // Second run (warm cache)
            //StandardLinq();
            //SamplePLinq();
            //SamplePLinq_NoLatency();
            //SamplePLinq_StopAndGo();
            //SamplePLinq_Inverted();
            //SamplePLinq_Ordered();
            //SamplePLinq_Unordered();
            //SamplePLinq_ExceptionHandling();
            //SamplePLinq_Cancellation();
            //SamplePLinq_ForceParallel();
            //SamplePLinq_Sequential();
            //SamplePLinq_RaceCondition();
            XmlPLinq();

        }

        static int[] LoadData(int size) {
            return Enumerable.Range(0, size - 1).ToArray(); 
        }
        
        //
        // Listing 16-18
        //
        static void StandardLinq() {
            Console.WriteLine("=== StandardLinq ===");
            int[] data = LoadData(100000000);
            var query =
                from i in data
                where i % 12345678 == 0
                select new { Value = i, ThreadID = Thread.CurrentThread.ManagedThreadId };

            Stopwatch sw = Stopwatch.StartNew();
            foreach (var number in query) {
                Console.WriteLine("{0} - from ThreadId={1}",
                                  number,
                                  Thread.CurrentThread.ManagedThreadId);
            }
            long elapsed = sw.ElapsedMilliseconds;
            Console.WriteLine("Elapsed time   : {0} milliseconds", elapsed);
            Console.WriteLine("CurrentThreadId: {0}", Thread.CurrentThread.ManagedThreadId);
        }

        //
        // Listing 16-19
        //
        static void SamplePLinq() {
            Console.WriteLine("=== SamplePLinq ===");
            int[] data = LoadData(100000000);
            var query =
                from i in data.AsParallel()
                where i % 12345678 == 0
                select new { Value = i, ThreadID = Thread.CurrentThread.ManagedThreadId };

            Stopwatch sw = Stopwatch.StartNew();
            foreach (var number in query) {
                Console.WriteLine("{0} - from ThreadId={1}",
                                  number,
                                  Thread.CurrentThread.ManagedThreadId);
            }
            long elapsed = sw.ElapsedMilliseconds;
            Console.WriteLine("Elapsed time   : {0} milliseconds", elapsed);
            Console.WriteLine("CurrentThreadId: {0}", Thread.CurrentThread.ManagedThreadId);
        }

        //
        // Listing 16-20
        //
        static void SamplePLinq_NoLatency() {
            Console.WriteLine("=== SamplePLinq_NoLatency ===");
            int[] data = LoadData(100000000);
            var query =
                from i in data.AsParallel().WithMergeOptions( ParallelMergeOptions.NotBuffered )
                where i % 12345678 == 0
                select new { Value = i, ThreadID = Thread.CurrentThread.ManagedThreadId, Timestamp = Stopwatch.GetTimestamp() };

            Stopwatch sw = Stopwatch.StartNew();
            foreach (var number in query) {
                long latency = Stopwatch.GetTimestamp() - number.Timestamp;
                Console.WriteLine("{0} - l={1} Th={2}",
                                    number,
                                    latency,
                                    Thread.CurrentThread.ManagedThreadId );
            }
            long elapsed = sw.ElapsedMilliseconds;
            Console.WriteLine("Elapsed time   : {0} milliseconds", elapsed);
            Console.WriteLine("CurrentThreadId: {0}", Thread.CurrentThread.ManagedThreadId);
        }

        //
        // Listing 16-21
        //
        static void SamplePLinq_StopAndGo() {
            Console.WriteLine("=== SamplePLinq_StopAndGo ===");
            int[] data = LoadData(100000000);
            var query =
                from i in data.AsParallel().WithMergeOptions( ParallelMergeOptions.FullyBuffered )
                where i % 12345678 == 0
                select new { Value = i, ThreadID = Thread.CurrentThread.ManagedThreadId, Timestamp = Stopwatch.GetTimestamp() };

            Stopwatch sw = Stopwatch.StartNew();
            foreach (var number in query) {
                long latency = Stopwatch.GetTimestamp() - number.Timestamp;
                Console.WriteLine("{0} - l={1} Th={2}",
                                    number,
                                    latency,
                                    Thread.CurrentThread.ManagedThreadId);
            }
            long elapsed = sw.ElapsedMilliseconds;
            Console.WriteLine("Elapsed time   : {0} milliseconds", elapsed);
            Console.WriteLine("CurrentThreadId: {0}", Thread.CurrentThread.ManagedThreadId);
        }

        //
        // Listing 16-22
        //
        static void SamplePLinq_Inverted() {
            Console.WriteLine("=== SamplePLinq_Inverted ===");
            int[] data = LoadData(100000000);
            var query =
                from i in data.AsParallel()
                where i % 12345678 == 0
                select new { Value = i, ThreadID = Thread.CurrentThread.ManagedThreadId };

            Stopwatch sw = Stopwatch.StartNew();
            query.ForAll((number) => {
                Console.WriteLine("{0} - from ThreadId={1}",
                                  number,
                                  Thread.CurrentThread.ManagedThreadId);
            });
            long elapsed = sw.ElapsedMilliseconds;
            Console.WriteLine("Elapsed time   : {0} milliseconds", elapsed);
            Console.WriteLine("CurrentThreadId: {0}", Thread.CurrentThread.ManagedThreadId);
        }

        //
        // Listing 16-23
        //
        static void SamplePLinq_Ordered() {
            Console.WriteLine("=== SamplePLinq_Ordered ===");
            int[] data = LoadData(100000000);
            var query =
                from i in data.AsParallel().AsOrdered()
                where i % 12345678 == 0
                select new { Value = i, ThreadID = Thread.CurrentThread.ManagedThreadId };
            
            Stopwatch sw = Stopwatch.StartNew();
            foreach (var number in query) {
                Console.WriteLine("{0} - ThreadId={1}",
                                    number,
                                    Thread.CurrentThread.ManagedThreadId);
            }
            long elapsed = sw.ElapsedMilliseconds;
            Console.WriteLine("Elapsed time   : {0} milliseconds", elapsed);
            Console.WriteLine("CurrentThreadId: {0}", Thread.CurrentThread.ManagedThreadId);
        }

        //
        // Listing 16-24
        //
        static void SamplePLinq_Unordered() {
            Console.WriteLine("=== SamplePLinq_Unordered ===");
            int[] data = LoadData(100000000);
            var queryEven =
                from i in data.AsParallel().AsOrdered()
                where i % 2 == 0
                select new { Value = i, ThreadID = Thread.CurrentThread.ManagedThreadId };

            var queryTop20Even = queryEven.Take(20);
            var queryFinal = from r in queryTop20Even.AsUnordered()
                                join s in queryTop20Even.AsUnordered()
                                on r.Value equals s.Value
                                where r.Value > 30
                                select r;

            Stopwatch sw = Stopwatch.StartNew();
            foreach (var number in queryFinal) {
                Console.WriteLine("{0} - ThreadId={1}",
                                    number,
                                    Thread.CurrentThread.ManagedThreadId);
            }
            long elapsed = sw.ElapsedMilliseconds;
            Console.WriteLine("Elapsed time   : {0} milliseconds", elapsed);
            Console.WriteLine("CurrentThreadId: {0}", Thread.CurrentThread.ManagedThreadId);
        }

        //
        // Listing 16-25
        //
        static void SamplePLinq_ExceptionHandling() {
            int[] data = LoadData(100);
            var query =
                from i in data.AsParallel()
                where i % 5 == 0
                select new {
                    Value = i * 2 / (i % 10),
                    ThreadID = Thread.CurrentThread.ManagedThreadId
                };

            try {
                foreach (var number in query) {
                    Console.WriteLine("{0} - from ThreadId={1}",
                                      number,
                                      Thread.CurrentThread.ManagedThreadId);
                }
            }
            catch (AggregateException ex) {
                Console.WriteLine(ex.Message);
                Console.WriteLine(
                    String.Format("AggregateException with {0} inner exceptions",
                                   ex.InnerExceptions.Count));
                int counter = 0;
                foreach (var innerException in ex.InnerExceptions) {
                    Console.WriteLine("({0})--> {1} : {2}",
                                      ++counter,
                                      innerException.GetType().Name,
                                      innerException.Message);
                }
            }
        }

        //
        // Listing 16-26
        //
        static void SamplePLinq_Cancellation() {
            Console.WriteLine("=== SamplePLinq_Cancellation ===");
            int[] data = LoadData(100000000);
            CancellationTokenSource cs = new CancellationTokenSource();

            // Requests an async cancellation in 100 milliseconds
            Task.Factory.StartNew(() => {
                Thread.Sleep(100);
                cs.Cancel();
            });

            var query =
                from i in data.AsParallel().WithCancellation( cs.Token )
                where i % 12345678 == 0
                select new { Value = i, ThreadID = Thread.CurrentThread.ManagedThreadId };

            try {
                foreach (var number in query) {
                    Console.WriteLine("{0} - from ThreadId={1}",
                                        number,
                                        Thread.CurrentThread.ManagedThreadId);
                }
                Console.WriteLine("CurrentThreadId: {0}", Thread.CurrentThread.ManagedThreadId);
            }
            catch (OperationCanceledException ex) {
                Console.WriteLine("Operation canceled: {0}", ex.Message);
            }
            catch (AggregateException ex) {
                if (ex.InnerExceptions != null) {
                    foreach (Exception e in ex.InnerExceptions) {
                        Console.WriteLine("AggregateException - {0}", e.Message);
                    }
                }
            }
        }

        static void SamplePLinq_ForceParallel() {
            Console.WriteLine("=== SamplePLinq_ForceParallel ===");
            int[] data = LoadData(10);
            CancellationTokenSource cs = new CancellationTokenSource();

            var query =
                from i in data.AsParallel().WithExecutionMode( ParallelExecutionMode.ForceParallelism )
                where i % 2 == 0
                select new { Value = i, ThreadID = Thread.CurrentThread.ManagedThreadId };

            try {
                foreach (var number in query) {
                    Console.WriteLine("{0} - from ThreadId={1}",
                                        number,
                                        Thread.CurrentThread.ManagedThreadId);
                }
                Console.WriteLine("CurrentThreadId: {0}", Thread.CurrentThread.ManagedThreadId);
            }
            catch (OperationCanceledException ex) {
                Console.WriteLine("Operation canceled: {0}", ex.Message);
            }
            catch (AggregateException ex) {
                if (ex.InnerExceptions != null) {
                    foreach (Exception e in ex.InnerExceptions) {
                        Console.WriteLine("AggregateException - {0}", e.Message);
                    }
                }
            }
        }

        static void SamplePLinq_Sequential() {
            Console.WriteLine("=== SamplePLinq_Sequential ===");
            int[] data = LoadData(100000000);

            var queryParallel =
                from i in data.AsParallel()
                where i % 12345678 == 0
                orderby i
                select i;

            var querySequential =
                from n in queryParallel.AsSequential().Take(3)
                select n;

            Stopwatch sw = Stopwatch.StartNew();
            foreach (var number in querySequential) {
                Console.WriteLine("{0} - from ThreadId={1}",
                                    number,
                                    Thread.CurrentThread.ManagedThreadId);
            }
            long elapsed = sw.ElapsedMilliseconds;
            Console.WriteLine("Elapsed time   : {0} milliseconds", elapsed);
            Console.WriteLine("CurrentThreadId: {0}", Thread.CurrentThread.ManagedThreadId);
        }

        static void SamplePLinq_RaceCondition() {
            Console.WriteLine("=== SamplePLinq_RaceCondition ===");
            int[] data = LoadData(10);

            int accumulator = 0;
            var query =
                from i in data.AsParallel()
                select new { Aggregated = accumulator += i, Value = i, ThreadID = Thread.CurrentThread.ManagedThreadId };
            
            Stopwatch sw = Stopwatch.StartNew();
            foreach (var number in query) {
                Console.WriteLine("{0} - from ThreadId={1}",
                                    number,
                                    Thread.CurrentThread.ManagedThreadId);
            }
            long elapsed = sw.ElapsedMilliseconds;
            Console.WriteLine("Elapsed time   : {0} milliseconds", elapsed);
            Console.WriteLine("CurrentThreadId: {0}", Thread.CurrentThread.ManagedThreadId);
            Console.WriteLine("Accumulator    : {0}", accumulator);
        }

        //
        // Listing 16-27
        //
        static void XmlPLinq() {
            XElement doc = XElement.Load("NorthwindCustomersWithOrders.xml");
            var query =
                from c in doc.Elements("customer").AsParallel()
                where c.Element("country").Value == "Italy"
                select new {
                    CompanyName = c.Element("companyName").Value,
                    ThreadID = Thread.CurrentThread.ManagedThreadId
                };

            foreach (var item in query) {
                Console.WriteLine("{0} - from ThreadId={1}",
                                  item,
                                  Thread.CurrentThread.ManagedThreadId);
            }
        }

    }
}
