﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;
using System.Text;
using System.Diagnostics;
using System.Threading;
using System.Threading.Tasks;
using System.Net.NetworkInformation;

namespace TPL {
    class Program {
        static void Main(string[] args) {
            // First run (cold cache)
            //DemoFor();
            //DemoForEach();
            //DemoInvoke();
            //DemoTask();
            //DemoTaskResult();
            //DemoContinuation();
            //DemoContinuationWhenAll();
            //DemoContinuationWhenAny();
            //DemoNestedTask();
            //DemoChildTask();
            //DemoCancelTask();
            //DemoExceptionTasks();
            //DemoExceptionChildTasks();
            //DemoApmCall();
            //DemoApmImplementation();
            //DemoEap();
            //DemoConcurrency_RaceCondition();
            DemoConcurrency_Safe();

            //// Second run (warm cache)
            //DemoFor();
            //DemoForEach();
            //DemoInvoke();
            //DemoTask();
            //DemoTaskResult();
            // DemoContinuation();
            //DemoContinuationWhenAll();
            //DemoContinuationWhenAny();
            //DemoNestedTask();
            //DemoChildTask();
            //DemoCancelTask();
            //DemoExceptionTasks();
            //DemoExceptionChildTasks();
            //DemoConcurrency_RaceCondition();
            //DemoConcurrency_Safe();
            //DemoConcurrency_RaceCondition();
            DemoConcurrency_Safe();
        }
        static void DemoFor() {
            SequentialFor();
            ParallelFor();
        }
        static int[] LoadData(int size) {
            int[] data = new int[size];
            for (int i = 0; i < data.Length; i++) {
                data[i] = i;
            }
            return data;
        }
        static void DemoForEach() {
            int[] data = LoadData(100);
            SequentialForEach(data);
            ParallelForEach(data);
        }
        static void ProcessData(int index) {
            // Simulate some data processing here
            for (long i = 0; i < 3000000 + index * 1000; i++) {
                ProcessItem(ref i);
            }
        }
        static void DataCalculation(int index) {
            // Simulate some data processing here
            for (long i = 0; i < 3000000 / (index + 1); i++) {
                ProcessItem(ref i);
            }
        }
        static void ProcessItem(ref long value) {
            value += 1;
            value -= 1;
        }

        static void DemoInvoke() {
            DemoInvoke_Sequential();
            DemoInvoke_Parallel();
        }


        //
        // Listing 16-1
        //
        static void SequentialFor() {
            Stopwatch sw = Stopwatch.StartNew();
            for (int index = 0; index < 100; index++) {
                ProcessData(index);
            }
            long elapsed = sw.ElapsedMilliseconds;
            Console.WriteLine("Sequential for: {0} milliseconds", elapsed);
        }

        static void ParallelFor() {
            Stopwatch sw = Stopwatch.StartNew();
            Parallel.For(0, 100, (index) => {
                ProcessData(index);
            });
            long elapsed = sw.ElapsedMilliseconds;
            Console.WriteLine("Parallel for: {0} milliseconds", elapsed);
        }

        //
        // Listing 16-2
        //
        static void SequentialForEach(int[] data) {
            Stopwatch sw = Stopwatch.StartNew();
            foreach (int value in data) {
                ProcessData(value);
            }
            long elapsed = sw.ElapsedMilliseconds;
            Console.WriteLine("Sequential foreach: {0} milliseconds", elapsed);
        }

        static void ParallelForEach(int[] data) {
            Stopwatch sw = Stopwatch.StartNew();
            Parallel.ForEach(data, (value) => {
                ProcessData(value);
            });
            long elapsed = sw.ElapsedMilliseconds;
            Console.WriteLine("Parallel foreach: {0} milliseconds", elapsed);
        }

        //
        // Listing 16-3
        //
        static int Operation(int index) {
            DataCalculation(index);
            Console.WriteLine("Operation {0}", index);
            return index * 10;
        }
        static void DemoInvoke_Sequential() {
            Console.WriteLine("=== Sequential calls ===");
            Operation(1);
            Operation(2);
            Operation(3);
            Operation(4);
        }
        static void DemoInvoke_Parallel() {
            Console.WriteLine("=== Parallel.Invoke calls ===");
            Parallel.Invoke(
                () => Operation(1),
                () => Operation(2),
                () => Operation(3),
                () => Operation(4));
        }


        //
        // Listing 16-4
        //
        static void DemoTask() {
            Console.WriteLine("=== Task calls ===");
            Task t1 = Task.Factory.StartNew(delegate { Operation(1); });
            Task t2 = Task.Factory.StartNew(delegate { Operation(2); });
            Task t3 = Task.Factory.StartNew(delegate { Operation(3); });
            Task t4 = Task.Factory.StartNew(delegate { Operation(4); });
            Task.WaitAll(new Task[] { t1, t2, t3, t4 });
        }

        //
        // Listing 16-5
        //
        static void DemoTaskResult() {
            Console.WriteLine("=== Task<TResult> calls ===");
            var f1 = Task.Factory.StartNew(() => Operation(1));
            var f2 = Task.Factory.StartNew(() => Operation(2));
            var f3 = Task.Factory.StartNew(() => Operation(3));
            var f4 = Task.Factory.StartNew(() => Operation(4));

            //Task.WaitAll(new Task[] { f1, f2, f3, f4 });
            int result = f1.Result + f2.Result + f3.Result + f4.Result;
            Console.WriteLine("Task result is {0}", result);
        }

        //
        // Listing 16-6
        //
        static void DemoContinuation() {
            Console.WriteLine("=== Task calls ===");
            Task tLast = Task.Factory.StartNew(() => { Operation(1); })
                .ContinueWith((prevTask) => { Operation(2); })
                .ContinueWith((prevTask) => { Operation(3); })
                .ContinueWith((prevTask) => { Operation(4); });
            tLast.Wait();
        }

        //
        // Listing 16-7
        //
        static void DemoContinuationWhenAll() {
            Console.WriteLine("=== Task calls ===");
            Task t1 = Task.Factory.StartNew(() => { Operation(1); });
            Task t2 = Task.Factory.StartNew(() => { Operation(2); });
            Task t3 = Task.Factory.StartNew(() => { Operation(3); });
            Task tLast = Task.Factory.ContinueWhenAll( 
                new Task[] { t1, t2, t3 },
                (prevTasks) => { 
                    Console.WriteLine("----"); 
                    Operation(4); });

            tLast.Wait();
        }

        //
        // No listing in the book
        //
        static void DemoContinuationWhenAny() {
            Console.WriteLine("=== Task calls ===");
            Task t1 = Task.Factory.StartNew(() => { Operation(1); });
            Task t2 = Task.Factory.StartNew(() => { Operation(2); });
            Task t3 = Task.Factory.StartNew(() => { Operation(3); });
            Console.WriteLine("Task ID t1={0}, t2={1}, t3={2}", t1.Id, t2.Id, t3.Id);
            Task tLast = Task.Factory.ContinueWhenAny(
                new Task[] { t1, t2, t3 },
                (prevTask) => {
                    Console.WriteLine("-- waited task={0}",prevTask.Id);
                    Operation(4);
                });

            tLast.Wait();
        }

        //
        // Listing 16-8
        //
        static void DemoNestedTask() {
            Console.WriteLine("=== Task calls ===");
            Task outerTask = Task.Factory.StartNew(() => { 
                Console.WriteLine("Outer task start");
                var nestedTask = Task.Factory.StartNew(() => { 
                    Console.WriteLine("Nested task start");
                    Console.WriteLine("Nested task complete");
                    return 42;
                });
                // If there was no call on nestedTask.Result, outerTask might end
                // before nestedTask completion
                Console.WriteLine("Nested result={0}", nestedTask.Result);
                Console.WriteLine("Outer task complete");
            });
            outerTask.Wait();
        }

        //
        // Listing 16-9
        //
        static void DemoChildTask() {
            Console.WriteLine("=== Task calls ===");
            Task parentTask = Task.Factory.StartNew(() => {
                Console.WriteLine("Parent task start");
                var childTask = Task.Factory.StartNew(
                    () => {
                        Console.WriteLine("Child task start");
                        Thread.Sleep(100);
                        Console.WriteLine("Child task complete");
                    },
                    TaskCreationOptions.AttachedToParent);
                Console.WriteLine("Parent task complete");
            });
            Console.WriteLine("Wait for outer task");
            parentTask.Wait();
            Console.WriteLine("Wait passed");
        }

        //
        // Listing 16-10
        //
        static void DemoCancelTask() {
            var tokenSource = new CancellationTokenSource();
            CancellationToken ct = tokenSource.Token;

            var task = Task.Factory.StartNew(() => {

                // Stop task if there is a pending cancellation request
                ct.ThrowIfCancellationRequested();

                bool moreToDo = true;
                while (moreToDo) {
                    // Check pending cancellation request (not necessary if cleanup code was not required)
                    if (ct.IsCancellationRequested) {
                        // Add cleanup code before stopping task

                        // Stop task (there is surely a cancellation request at this point)
                        ct.ThrowIfCancellationRequested();
                    }

                    Thread.Sleep(100);
                    Console.Write(".");
                }
            }, tokenSource.Token); // Pass same token to StartNew.

            // Wait for ENTER, then cancel execution
            Console.ReadLine();
            tokenSource.Cancel();

            // Just continue on this thread, or Wait/WaitAll with try-catch:
            try {
                task.Wait();
            }
            catch (AggregateException e) {
                foreach (var v in e.InnerExceptions) {
                    Console.WriteLine("{0}: {1}", e.Message, v.Message);
                }
            }
        }

        //
        // Listing 16-11
        //
        static void DemoExceptionTasks() {
            try {
                Parallel.Invoke(
                    () => { throw new ArgumentException(); },
                    () => { throw new NullReferenceException(); }
                );
            }
            catch (AggregateException ex) {
                foreach (var v in ex.InnerExceptions) {
                    Console.WriteLine("Exception: " + v.Message);
                }
            }
        }

        //
        // Listing 16-12
        //
        static void DemoExceptionChildTasks() {
            Console.WriteLine("=== Task calls ===");
            Task parentTask = Task.Factory.StartNew(() => {
                Console.WriteLine("Parent task start");
                var childTask = Task.Factory.StartNew(
                    () => {
                        Console.WriteLine("Child task start");
                        Thread.Sleep(100);
                        Console.WriteLine("Child task complete");
                        throw new ArgumentException();
                    },
                    TaskCreationOptions.AttachedToParent);
                Console.WriteLine("Parent task complete");
            });
            Console.WriteLine("Wait for outer task");
 
            try {
                parentTask.Wait();
                Console.WriteLine("Wait passed");

                Thread.Sleep(200);
            }
            catch (AggregateException ex) {
                foreach (var v in ex.InnerExceptions) {
                    Console.WriteLine("{0}: {1}", v.GetType().Name, v.Message);

                    // Exceptions from child tasks are in another AggregateException
                    if (v is AggregateException) {
                        foreach (var childEx in (v as AggregateException).InnerExceptions) {
                            Console.WriteLine("Child {0}: {1}", childEx.GetType().Name, childEx.Message);
                        }
                    }
                }
            }
        }

        //
        // Listing 16-13
        //
        static void DemoApmCall() {
            // We create two tasks to copy file. The first read the file, the second write the content in another file
            string pathRead = "SampleData.txt";
            string pathWrite = "CopyData.txt";

            FileInfo fi = new FileInfo(pathRead);
            byte[] buffer = new byte[fi.Length];

            FileStream fr = new FileStream(pathRead, FileMode.Open, FileAccess.Read, FileShare.Read, buffer.Length, true);

            // taskRead will return the number of bytes read (it is the same type returned by EndRead)
            Task<int> taskRead = Task<int>.Factory.FromAsync(fr.BeginRead, fr.EndRead, buffer, 0, buffer.Length, null);
            Task taskCopy = taskRead.ContinueWith((tr) => {
                // Close source file
                fr.Close();

                // Check if read length is correct
                if (tr.Result != buffer.Length) {
                    throw new Exception("Error reading source file");
                }

                FileStream fw = new FileStream(pathWrite, FileMode.Create, FileAccess.Write, FileShare.None, buffer.Length, true);
                
                // EndWrite is void - so Task is required
                Task taskWrite = Task.Factory.FromAsync(fw.BeginWrite, fw.EndWrite, buffer, 0, buffer.Length, TaskCreationOptions.AttachedToParent)
                                 .ContinueWith((tw) => {
                                     fw.Close();
                                  }, TaskContinuationOptions.AttachedToParent);
            });

            try {
                taskCopy.Wait();
            }
            catch (AggregateException ex) {
                foreach (var v in ex.InnerExceptions) {
                    Console.WriteLine("{0}:{1}", v.GetType().Name, v.Message);
                }
            }
        }


        // 
        // Listing 16-14
        //
        class Calculator {
            public IAsyncResult BeginCalculate(int calcParameter, AsyncCallback ac, object state) {
                Console.WriteLine("Calling BeginCalculate on thread {0}", Thread.CurrentThread.ManagedThreadId);

                Task<int> f = Task.Factory.StartNew( (s) => Compute(calcParameter), state);
                if (ac != null) f.ContinueWith((res) => ac(f));
                return f;
            }

            public int Compute(int parameter) {
                Console.WriteLine("Calling compute on thread {0}", Thread.CurrentThread.ManagedThreadId);

                return parameter * 42;
            }

            public int EndCalculate(IAsyncResult ar) {
                Console.WriteLine("Calling EndCalculate on thread {0}", Thread.CurrentThread.ManagedThreadId);

                return ((Task<int>)ar).Result;
            }
        }

        static void DemoApmImplementation() {
            Calculator calc = new Calculator();
            calc.BeginCalculate(1, (ar) => {
                Console.WriteLine(calc.EndCalculate(ar));
            }, null);
            Console.ReadLine();

        }

        // 
        // Listing 16-15
        //
        static Task<long[]> PingTask(string url, int packets, CancellationToken token) {
            TaskCompletionSource<long[]> tcs = new TaskCompletionSource<long[]>();
            Ping[] pings = new Ping[packets];
            long[] timings = new long[packets];
            int count = 0;

            // In case of cancellation, send cancel to all active pings
            token.Register(() => {
                foreach (var ping in pings) {
                    if (ping != null) {
                        ping.SendAsyncCancel();
                    }
                }
            });

            for (int i = 0; i < packets; i++) {
                // Saves i in currentPing - it will be saved in the PingCompleted asynchronous event handler
                int currentPing = i;
                pings[currentPing] = new Ping();
                Ping ping = pings[currentPing];
                ping.PingCompleted += new PingCompletedEventHandler((obj, args) => {
                    
                    // BUG IF YOU WRITE THIS: timings[i] = ...
                    timings[currentPing] = (args.Reply != null) ? args.Reply.RoundtripTime : -1;
                    // Check if it is the last operation
                    if (Interlocked.Increment(ref count) == packets) {
                        // if last ping completed, set Result on underlying task
                        tcs.TrySetResult(timings);
                    }
                });
                ping.SendAsync(url, null);
            }
            return tcs.Task;
        }

        static void DemoEap() {
            string url = "www.bing.com";
            CancellationTokenSource ts = new CancellationTokenSource();
            var pingsTask = PingTask(url, 5, ts.Token);

            // Call ts.Cancel() if you want to cancel the pingsTask
            
            long[] timings = pingsTask.Result;
            foreach (long t in timings) {
                Console.WriteLine("Time: {0}", t);
            }
        }

        // 
        // Listing 16-16
        //
        static void DemoConcurrency_RaceCondition() {
            Console.WriteLine("=== DemoConcurrency_RaceCondition ===");
            int sum = 0;
            Parallel.For(1, 10, (index) => {
                int local = sum;
                local += Operation(index);
                sum = local;
            });
            Console.WriteLine("Final sum: {0}", sum);
        }

        // 
        // Listing 16-17
        //
        static void DemoConcurrency_Safe() {
            Console.WriteLine("=== DemoConcurrency_Safe ===");
            int sum = 0;
            int[] results = new int[10];
            Parallel.For(1, 10, (index) => {
                results[index] = Operation(index);
            });
            sum = results.Sum();
            Console.WriteLine("Final sum: {0}", sum);
        }

    }
}
