﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Collections;

namespace DevLeap.Linq.Syntax
{
    #region Supporting types

    public class Developer
    {
        public string Name;
        public string Language;
        public int Age;
    }

    public class Customer
    {
        public String Name { get; set; }
        public String City { get; set; }
        public Order[] Orders { get; set; }
    }

    public class Order
    {
        public Int32 IdOrder { get; set; }
        public Decimal EuroAmount { get; set; }
        public String Description { get; set; }
    }

    public class Category
    {
        public Int32 IdCategory { get; set; }
        public String Name { get; set; }
    }

    public class Product
    {
        public String IdProduct { get; set; }
        public Int32 IdCategory { get; set; }
        public String Description { get; set; }
    }

    #endregion

    class App
    {
        static void Main(string[] args)
        {
            //getCSharpDevelopersNames();
            //getCSharpDevelopersNamesDeconstructed();
            //getCSharpDevelopersNamesWithCast();
            //queryWithJoin();
            //queryWithWhereClause();
            //getDevelopersNamesByLanguage();
            //getDevelopersNamesByLanguageAndAge();
            //getDevelopersCountGroupedByLanguage();
            //queryWithOrderBy();
            //queryWithOrderByMultipleOrderingFields();
            //innerJoinQuery();
            //groupJoinQuery();
            //leftOuterJoinQuery();
            //categoriesByProductsNumber();
            //deferredQueryEvaluation();
            //extensionMethodResolution();
            //degenerateQueryExpression();
            //directIteration();
            //queryWithException();
            //queryWithExceptionHandled();
            //queryWithExceptionHandledInDataSourceDefinition();

            Console.ReadLine();
        }

        //
        // Listing 2-1
        //
        static void getCSharpDevelopersNames()
        {
            Developer[] developers = new Developer[] {
                new Developer {Name = "Paolo", Language = "C#"},
                new Developer {Name = "Marco", Language = "C#"},
                new Developer {Name = "Frank", Language = "VB.NET"}};

            IEnumerable<string> developersUsingCSharp =
                from d in developers
                where d.Language == "C#"
                select d.Name;

            foreach (string item in developersUsingCSharp)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 2-3
        //
        static void getCSharpDevelopersNamesDeconstructed()
        {
            Developer[] developers = new Developer[] {
                new Developer {Name = "Paolo", Language = "C#"},
                new Developer {Name = "Marco", Language = "C#"},
                new Developer {Name = "Frank", Language = "VB.NET"}};

            Func<Developer, bool> filteringPredicate = d => d.Language == "C#";
            Func<Developer, string> selectionPredicate = d => d.Name;
            IEnumerable<string> expr =
                developers
                .Where(filteringPredicate)
                .Select(selectionPredicate);

            foreach (string item in expr)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 2-4
        //
        static void getCSharpDevelopersNamesWithCast()
        {
            ArrayList developers = new ArrayList();
            developers.Add(new Developer { Name = "Paolo", Language = "C#" });
            developers.Add(new Developer { Name = "Marco", Language = "C#" });
            developers.Add(new Developer { Name = "Frank", Language = "VB.NET" });

            IEnumerable<string> developersUsingCSharp =
                from Developer d in developers
                where d.Language == "C#"
                select d.Name;

            foreach (string item in developersUsingCSharp)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 2-5
        //
        static void queryWithJoin()
        {
            Customer[] customers = new Customer[] {
                new Customer { Name = "Paolo", City = "Brescia", 
                    Orders = new Order[] {
                        new Order { IdOrder = 1, EuroAmount = 100, Description = "Order 1" },
                        new Order { IdOrder = 2, EuroAmount = 150, Description = "Order 2" },
                        new Order { IdOrder = 3, EuroAmount = 230, Description = "Order 3" },
                    }},
                new Customer { Name = "Marco", City = "Torino", 
                    Orders = new Order[] {
                        new Order { IdOrder = 4, EuroAmount = 320, Description = "Order 4" },
                        new Order { IdOrder = 5, EuroAmount = 170, Description = "Order 5" },
                    }}};

            var ordersQuery =
                from c in customers
                from o in c.Orders
                select new { c.Name, o.IdOrder, o.EuroAmount };

            foreach (var item in ordersQuery)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 2-7
        //
        static void queryWithWhereClause()
        {
            Customer[] customers = new Customer[] {
                new Customer { Name = "Paolo", City = "Brescia", 
                    Orders = new Order[] {
                        new Order { IdOrder = 1, EuroAmount = 100, Description = "Order 1" },
                        new Order { IdOrder = 2, EuroAmount = 150, Description = "Order 2" },
                        new Order { IdOrder = 3, EuroAmount = 230, Description = "Order 3" },
                    }},
                new Customer { Name = "Marco", City = "Torino", 
                    Orders = new Order[] {
                        new Order { IdOrder = 4, EuroAmount = 320, Description = "Order 4" },
                        new Order { IdOrder = 5, EuroAmount = 170, Description = "Order 5" },
                    }}};

            var ordersQuery =
                from c in customers
                from o in c.Orders
                where o.EuroAmount > 200
                select new { c.Name, o.IdOrder, o.EuroAmount };

            foreach (var item in ordersQuery)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 2-9
        //
        static void getDevelopersNamesByLanguage()
        {
            Developer[] developers = new Developer[] {
                new Developer { Name = "Paolo", Language = "C#" },
                new Developer { Name = "Marco", Language = "C#" },
                new Developer { Name = "Frank", Language = "VB.NET" },
            };

            var developersGroupedByLanguage =
                from d in developers
                group d by d.Language;

            foreach (var group in developersGroupedByLanguage)
            {
                Console.WriteLine("Language: {0}", group.Key);
                foreach (var item in group)
                {
                    Console.WriteLine("\t{0}", item.Name);
                }
            }
        }

        //
        // Listing 2-10
        //
        static void getDevelopersNamesByLanguageAndAge()
        {
            Developer[] developers = new Developer[] {
                new Developer { Name = "Paolo", Language = "C#", Age = 32 },
                new Developer { Name = "Marco", Language = "C#", Age = 37},
                new Developer { Name = "Frank", Language = "VB.NET", Age = 48  },
            };

            var developersGroupedByLanguage =
                from d in developers
                group d by new { d.Language, AgeCluster = (d.Age / 10) * 10 };

            foreach (var group in developersGroupedByLanguage)
            {
                Console.WriteLine("Language: {0}", group.Key);
                foreach (var item in group)
                {
                    Console.WriteLine("\t{0}", item.Name);
                }
            }
        }

        //
        // Listing 2-12
        //
        static void getDevelopersCountGroupedByLanguage()
        {
            Developer[] developers = new Developer[] {
                new Developer { Name = "Paolo", Language = "C#", Age = 32 },
                new Developer { Name = "Marco", Language = "C#", Age = 37},
                new Developer { Name = "Frank", Language = "VB.NET", Age = 48  },
            };

            var developersGroupedByLanguage =
                from d in developers
                group d by d.Language into developersGrouped
                select new { Language = developersGrouped.Key, DevelopersCount = developersGrouped.Count() };

            foreach (var group in developersGroupedByLanguage)
            {
                Console.WriteLine("Language {0} contains {1} developers", group.Language, group.DevelopersCount);
            }
        }


        //
        // Listing 2-13
        //
        static void queryWithOrderBy()
        {
            Customer[] customers = new Customer[] {
                new Customer { Name = "Paolo", City = "Brescia", 
                    Orders = new Order[] {
                        new Order { IdOrder = 1, EuroAmount = 100, Description = "Order 1" },
                        new Order { IdOrder = 2, EuroAmount = 150, Description = "Order 2" },
                        new Order { IdOrder = 3, EuroAmount = 230, Description = "Order 3" },
                    }},
                new Customer { Name = "Marco", City = "Torino", 
                    Orders = new Order[] {
                        new Order { IdOrder = 4, EuroAmount = 320, Description = "Order 4" },
                        new Order { IdOrder = 5, EuroAmount = 170, Description = "Order 5" },
                    }}};

            var ordersSortedByEuroAmount =
                from c in customers
                from o in c.Orders
                orderby o.EuroAmount
                select new { c.Name, o.IdOrder, o.EuroAmount };

            foreach (var item in ordersSortedByEuroAmount)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 2-14
        //
        static void queryWithOrderByMultipleOrderingFields()
        {
            Customer[] customers = new Customer[] {
                new Customer { Name = "Paolo", City = "Brescia", 
                    Orders = new Order[] {
                        new Order { IdOrder = 1, EuroAmount = 100, Description = "Order 1" },
                        new Order { IdOrder = 2, EuroAmount = 150, Description = "Order 2" },
                        new Order { IdOrder = 3, EuroAmount = 230, Description = "Order 3" },
                    }},
                new Customer { Name = "Marco", City = "Torino", 
                    Orders = new Order[] {
                        new Order { IdOrder = 4, EuroAmount = 320, Description = "Order 4" },
                        new Order { IdOrder = 5, EuroAmount = 170, Description = "Order 5" },
                    }}};

            var ordersSortedByCustomerAndEuroAmount =
                from c in customers
                from o in c.Orders
                orderby c.Name, o.EuroAmount descending
                select new { c.Name, o.IdOrder, o.EuroAmount };

            foreach (var item in ordersSortedByCustomerAndEuroAmount)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 2-16
        //
        static void innerJoinQuery()
        {
            Category[] categories = new Category[] {
                new Category { IdCategory = 1, Name = "Pasta"},
                new Category { IdCategory = 2, Name = "Beverages"},
                new Category { IdCategory = 3, Name = "Other food"},
            };

            Product[] products = new Product[] {
                new Product { IdProduct = "PASTA01", IdCategory = 1, Description = "Tortellini" },
                new Product { IdProduct = "PASTA02", IdCategory = 1, Description = "Spaghetti" },
                new Product { IdProduct = "PASTA03", IdCategory = 1, Description = "Fusilli" },
                new Product { IdProduct = "BEV01", IdCategory = 2, Description = "Water" },
                new Product { IdProduct = "BEV02", IdCategory = 2, Description = "Orange Juice" },
                // Uncomment this products in order to have some product in "Other food" category
                // new Product { IdProduct = "FOOD01", IdCategory = 3, Description = "Cheese Cake" },
                // new Product { IdProduct = "FOOD02", IdCategory = 3, Description = "Biscuits" },
            };

            var categoriesAndProducts =
                from c in categories
                join p in products on c.IdCategory equals p.IdCategory
                select new { c.IdCategory, CategoryName = c.Name, Product = p.Description };

            foreach (var item in categoriesAndProducts)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 2-17
        //
        static void groupJoinQuery()
        {
            Category[] categories = new Category[] {
                new Category { IdCategory = 1, Name = "Pasta"},
                new Category { IdCategory = 2, Name = "Beverages"},
                new Category { IdCategory = 3, Name = "Other food"},
            };

            Product[] products = new Product[] {
                new Product { IdProduct = "PASTA01", IdCategory = 1, Description = "Tortellini" },
                new Product { IdProduct = "PASTA02", IdCategory = 1, Description = "Spaghetti" },
                new Product { IdProduct = "PASTA03", IdCategory = 1, Description = "Fusilli" },
                new Product { IdProduct = "BEV01", IdCategory = 2, Description = "Water" },
                new Product { IdProduct = "BEV02", IdCategory = 2, Description = "Orange Juice" },
            };

            var categoriesAndProducts =
                from c in categories
                join p in products on c.IdCategory equals p.IdCategory into productsByCategory
                select new { c.IdCategory, CategoryName = c.Name, Products = productsByCategory };

            foreach (var category in categoriesAndProducts)
            {
                Console.WriteLine("{0} - {1}", category.IdCategory, category.CategoryName);
                foreach (var product in category.Products)
                {
                    Console.WriteLine("\t{0}", product.Description);
                }
            }
        }

        //
        // Listing 2-18
        //
        static void leftOuterJoinQuery()
        {
            Category[] categories = new Category[] {
                new Category { IdCategory = 1, Name = "Pasta"},
                new Category { IdCategory = 2, Name = "Beverages"},
                new Category { IdCategory = 3, Name = "Other food"},
            };

            Product[] products = new Product[] {
                new Product { IdProduct = "PASTA01", IdCategory = 1, Description = "Tortellini" },
                new Product { IdProduct = "PASTA02", IdCategory = 1, Description = "Spaghetti" },
                new Product { IdProduct = "PASTA03", IdCategory = 1, Description = "Fusilli" },
                new Product { IdProduct = "BEV01", IdCategory = 2, Description = "Water" },
                new Product { IdProduct = "BEV02", IdCategory = 2, Description = "Orange Juice" },
            };

            var categoriesAndProducts =
                from c in categories
                join p in products on c.IdCategory equals p.IdCategory into productsByCategory
                from pc in productsByCategory.DefaultIfEmpty(new Product { IdProduct = String.Empty, Description = String.Empty, IdCategory = 0 })
                select new { c.IdCategory, CategoryName = c.Name, Product = pc.Description };

            foreach (var item in categoriesAndProducts)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 2-21
        //
        static void categoriesByProductsNumber()
        {
            Category[] categories = new Category[] {
                new Category { IdCategory = 1, Name = "Pasta"},
                new Category { IdCategory = 2, Name = "Beverages"},
                new Category { IdCategory = 3, Name = "Other food"},
            };

            Product[] products = new Product[] {
                new Product { IdProduct = "PASTA01", IdCategory = 1, Description = "Tortellini" },
                new Product { IdProduct = "PASTA02", IdCategory = 1, Description = "Spaghetti" },
                new Product { IdProduct = "PASTA03", IdCategory = 1, Description = "Fusilli" },
                new Product { IdProduct = "BEV01", IdCategory = 2, Description = "Water" },
                new Product { IdProduct = "BEV02", IdCategory = 2, Description = "Orange Juice" },
            };

            var categoriesByProductsNumberQuery =
                from c in categories
                join p in products on c.IdCategory equals p.IdCategory into productsByCategory
                let ProductsCount = productsByCategory.Count()
                orderby ProductsCount
                select new { c.IdCategory, ProductsCount };

            foreach (var item in categoriesByProductsNumberQuery)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 2-22
        // 
        static void deferredQueryEvaluation()
        {
            List<Developer> developers = new List<Developer>(new Developer[] {
                new Developer { Name = "Paolo", Language = "C#", Age = 32 },
                new Developer { Name = "Marco", Language = "C#", Age = 37},
                new Developer { Name = "Frank", Language = "VB.NET", Age = 48  },
            });

            var query =
                from d in developers
                where d.Language == "C#"
                select new { d.Name, d.Age };

            Console.WriteLine("There are {0} C# developers.", query.Count());

            // Listing 2-23
            developers.Add(new Developer { Name = "Roberto", Language = "C#", Age = 35 });
            Console.WriteLine("There are {0} C# developers.", query.Count());

        }

        //
        // Listing 2-25
        //
        static void extensionMethodResolution()
        {
            Developers developers = new Developers(new Developer[] {
                new Developer { Name = "Paolo", Language = "C#", Age = 32 },
                new Developer { Name = "Marco", Language = "C#", Age = 37},
                new Developer { Name = "Frank", Language = "VB.NET", Age = 48  },
            });

            var query =
                from d in developers
                where d.Language == "C#"
                select d;

            Console.WriteLine("There are {0} C# developers.", query.Count());
        }

        //
        // Listing 2-26
        //
        static void degenerateQueryExpression()
        {
            Developer[] developers = new Developer[] {
                new Developer { Name = "Paolo", Language = "C#", Age = 32 },
                new Developer { Name = "Marco", Language = "C#", Age = 37},
                new Developer { Name = "Frank", Language = "VB.NET", Age = 48 },
            };

            var query =
                from d in developers
                select d;

            foreach (var developer in query)
            {
                Console.WriteLine("{0} {1}", developer.Name, developer.GetHashCode());
            }
        }

        //
        // Listing 2-27
        //
        static void directIteration()
        {
            Developer[] developers = new Developer[] {
                new Developer { Name = "Paolo", Language = "C#", Age = 32 },
                new Developer { Name = "Marco", Language = "C#", Age = 37},
                new Developer { Name = "Frank", Language = "VB.NET", Age = 48 },
            };

            foreach (var developer in developers)
            {
                Console.WriteLine("{0} {1}", developer.Name, developer.GetHashCode());
            }
        }

        //
        // Listing 2-28
        //
        static Boolean DoSomething(Developer dev)
        {
            if (dev.Age > 40)
                throw new ArgumentOutOfRangeException("dev");

            return (dev.Language == "C#");
        }

        static void queryWithException()
        {
            Developer[] developers = new Developer[] {
                new Developer { Name = "Paolo", Language = "C#", Age = 32 },
                new Developer { Name = "Marco", Language = "C#", Age = 37},
                new Developer { Name = "Frank", Language = "VB.NET", Age = 48  },
            };

            var query =
                from d in developers
                let SomethingResult = DoSomething(d)
                select new { d.Name, SomethingResult };

            foreach (var item in query)
            {
                Console.WriteLine(item);
            }
        }

        //
        // Listing 2-29
        // 
        static void queryWithExceptionHandled()
        {
            Developer[] developers = new Developer[] {
                new Developer { Name = "Paolo", Language = "C#", Age = 32 },
                new Developer { Name = "Marco", Language = "C#", Age = 37},
                new Developer { Name = "Frank", Language = "VB.NET", Age = 48  },
            };

            var query =
                from d in developers
                let SomethingResult = DoSomething(d)
                select new { d.Name, SomethingResult };

            try
            {
                foreach (var item in query)
                {
                    Console.WriteLine(item);
                }
            }
            catch (ArgumentOutOfRangeException e)
            {
                Console.WriteLine(e.Message);
            }
        }

        //
        // Listing 2-30
        // 
        static void queryWithExceptionHandledInDataSourceDefinition()
        {
            Developer[] developers = null;

            try
            {
                developers = createDevelopersDataSource();
            }
            catch (InvalidOperationException e)
            {
                // Imagine that the createDevelopersDataSource
                // throws an InvalidOperationException in case of failure

                // Handle it somehow ...
                Console.WriteLine(e.Message);
            }

            if (developers != null)
            {
                var query =
                    from d in developers
                    let SomethingResult = DoSomething(d)
                    select new { d.Name, SomethingResult };

                try
                {
                    foreach (var item in query)
                    {
                        Console.WriteLine(item);
                    }
                }
                catch (ArgumentOutOfRangeException e)
                {
                    Console.WriteLine(e.Message);
                }
            }
        }

        private static Developer[] createDevelopersDataSource()
        {
            // Fictitious InvalidOperationException thrown
            throw new InvalidOperationException();
        }
    }
}
